# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from abc import ABC
from typing import Dict, Optional, List

import dash
from dash import dcc, html, Output, Input, State
import dash_bootstrap_components as dbc
from dash_bootstrap_templates import load_figure_template
import os
from flask_caching import Cache
import enum
import diskcache
from dash.long_callback import DiskcacheLongCallbackManager

from dse_do_dashboard.utils.dash_common_utils import ScenarioTableSchema


class HostEnvironment(enum.Enum):
    Local = 1  # Regular Dash
    CPD402 = 2  # Special handling of port


class DashApp(ABC):
    def __init__(self, logo_file_name: str = 'IBM.png',
                 navbar_brand_name: Optional[str] = 'Dashboard',
                 cache_config: Dict = {},
                 port: int = 8050,
                 dash_debug: bool = False,
                 host_env: Optional[HostEnvironment] = None,
                 bootstrap_theme = dbc.themes.BOOTSTRAP,
                 bootstrap_figure_template: str = "bootstrap",
                 enable_long_running_callbacks: bool = False,
                 ):
        self.port = port
        self.host_env = host_env
        self.dash_debug = dash_debug
        self.bootstrap_theme = bootstrap_theme
        self.set_bootstrap_figure_template(bootstrap_figure_template)

        # Long running callbacks:
        self.enable_long_running_callbacks = enable_long_running_callbacks
        if self.enable_long_running_callbacks:
            cache = diskcache.Cache("./cache")
            self.long_callback_manager = DiskcacheLongCallbackManager(cache)
        else:
            self.long_callback_manager = None

        self.app = self.create_dash_app()

        # Margins to layout the header, sidebar and content area:
        self.header_height = "3rem" # "6rem"
        self.footer_height = "10rem"
        self.sidebar_width = "16rem"
        self.adbar_width = "4rem"

        self.logo_file_name = logo_file_name  # set before call to get_app_layout()
        self.navbar_brand_name = navbar_brand_name
        self.cache_config = cache_config

        self.config_cache()
        self.app.layout = self.get_app_layout()

        self.set_cache_callbacks()
        self.set_dash_callbacks()

        # To run locally on Mac, you need to specifically set host='localhost'
        self.run_server_kwargs: Dict = {}  # {"host": "localhost"}

    def set_run_server_kwargs(self, **run_server_kwargs):
        """Use to set input arguments for Dash.run_server(), in addition to `debug` and `port`.
        For instance, running locally on a Mac requires host='localhost'. """
        if run_server_kwargs is not None:
            self.run_server_kwargs = run_server_kwargs

    def set_bootstrap_figure_template(self, bootstrap_figure_template: str):
        """See https://hellodash.pythonanywhere.com/theme_explorer"""
        load_figure_template(bootstrap_figure_template)
        self.bootstrap_figure_template = bootstrap_figure_template

    def create_dash_app(self):
        """Creates the Dash app. Called from the DashApp constructor.
        Implements special process around running in CPDv4.0.2 (i.e. `ws_applications.make_link()`)
        Can be overridden if necessary.

        Note:
            Changing Bootstrap stylesheets requires a Reload on the browser!
        """
        # print(os.getcwd())
        assets_path = os.getcwd() +'/assets'  # Not sure why it is not picking it up automatically, but this works
        if self.host_env == HostEnvironment.CPD402:
            print(f"Host is CPD, so run with make_link({self.port})")
            from ws_applications import make_link
            requests_prefix = make_link(self.port)
            app = dash.Dash(__name__,
                            serve_locally=False,
                            requests_pathname_prefix=requests_prefix,
                            # suppress_callback_exceptions = True,
                            assets_folder=assets_path,
                            long_callback_manager=self.long_callback_manager,
                            )
        else:
            app = dash.Dash(__name__,
                            # suppress_callback_exceptions = True,
                            assets_folder=assets_path,
                            long_callback_manager=self.long_callback_manager,
                            )
        dbc_css = "https://cdn.jsdelivr.net/gh/AnnMarieW/dash-bootstrap-templates/dbc.css"
        app.config.external_stylesheets = [self.bootstrap_theme, dbc_css]
        # app.config.external_stylesheets = [dbc.themes.SOLAR]
        app.config.suppress_callback_exceptions = True
        return app

    def run_server(self):
        """Runs the Dash server.
        To be called from index.py::

            if __name__ == '__main__':
                DA.run_server()

        """
        self.app.run_server(debug=self.dash_debug, port=self.port, **self.run_server_kwargs)

    def config_cache(self):
        self.cache = Cache()
        self.cache.init_app(self.app.server, config=self.cache_config)

    def get_app_layout(self):
        """Layout of whole app"""
        layout = html.Div([
            dcc.Location(id='url'),
            self.get_app_stores_div(),
            self.get_navbar(),
            self.get_sidebar(),
            self.get_content_template()],
            className="dbc"  # Bootstrap CSS, see https://hellodash.pythonanywhere.com/about_dbc_css
        )
        return layout

    def display_content_callback(self, pathname):
        """Callback for main content area. Will update the content area.
        Will need to be called through callback in index.py, where `DA` is the instance of the DashApp.

        Usage::

            @app.callback(Output('content', 'children'), [Input('url', 'pathname')])
            def display_content(pathname):
                return DA.display_content_callback()

        """
        layout = html.Div([
            html.H1(f"Error 404 - Page '{pathname}' not found"),
            dcc.Textarea(
                id='not_found',
                value='Page not found',
                style={'width': '100%', 'height': 300},
            ),
        ])
        return layout

    def get_content_template(self):
        """The template structure for the main content page.
        The callback `display_content` will update the actual contents
        """
        content = html.Div(id = 'content',
                           style = self.get_content_style()
                           )
        return content

    def get_content_style(self):
        """Style for the main contents section.
        Very important is the `margin-left`, otherwise it will overlap with the left-side navbar.
        Except for the margins, the other properties don't seem to differ from the default.

        :return:
        """
        # header_height, footer_height = "6rem", "10rem"
        # sidebar_width, adbar_width = "16rem", "4rem"
        content_style = {
            "margin-top": "1rem" , #self.header_height,
            "margin-left": self.sidebar_width,
            "margin-right": "1rem" , #self.adbar_width,
            "margin-bottom": self.footer_height,
            # "padding": "2rem 1rem",
            # "card_border": {
            #     "width": "1px",
            #     "style": "solid",
            #     "color": "#8EA9C1",
            #     "radius": "0px"
            # },
            # "card_background_color": "#edf3f4",
            # "card_box_shadow": "0px 0px 0px rgba(0,0,0,0)",
            # "card_margin": "15px",
            # "card_padding": "5px",
            # "card_outline": {
            #     "width": "0px",
            #     "style": "solid",
            #     "color": "#8EA9C1"
            # },
            # "card_header_border": {
            #     "width": "0px 0px 2px 0px",
            #     "style": "dashed",
            #     "color": "#8EA9C1",
            #     "radius": "0px"
            # },
            # "card_header_background_color": "#edf3f4",
            # "card_header_box_shadow": "0px 0px 0px rgba(0,0,0,0)",
            # "card_header_margin": "0px",
            # "card_header_padding": "10px",
            # "colorway": [
            #     "#004172",
            #     "#08539d",
            #     "#2e64c7",
            #     "#be35a0",
            #     "#e32433",
            #     "#eb6007",
            #     "#fb8b00",
            #     "#c19f00",
            #     "#5c9c00",
            #     "#897500",
            #     "#cb0049",
            #     "#7746ba",
            #     "#0080d1",
            #     "#3192d2",
            #     "#ac6ac0",
            #     "#e34862",
            #     "#c57e00",
            #     "#71a500",
            #     "#ad6e00",
            #     "#b82e2e",
            # ],
            # "colorscale": [
            #     "#00224e",
            #     "#123570",
            #     "#3b496c",
            #     "#575d6d",
            #     "#707173",
            #     "#8a8678",
            #     "#a59c74",
            #     "#c3b369",
            #     "#e1cc55",
            #     "#fee838"
            # ],
            # "font_family": "Quicksand",
            # "font_size": "15px",
            # "font_size_smaller_screen": "13px",
            # "font_family_header": "PT Sans",
            # "font_size_header": "24px",
            # "font_family_headings": "PT Sans",
            # "font_headings_size": None,
            # "header_border": {
            #     "width": "0px",
            #     "style": "solid",
            #     "color": "#8EA9C1",
            #     "radius": "0px"
            # },
        }
        return content_style

    def get_logo_src(self):
        src = self.app.get_asset_url(self.logo_file_name)
        return src

    def get_navbar_brand_children(self) -> str:
        """The children of the dbc.NavbarBrand component, typically a plain string.
        To be overridden."""
        brand = self.navbar_brand_name  #'Dashboard'
        return brand

    def get_navbar(self):
        # The scenario-bar is the section on the right with the scenario-dropdown and the refresh button
        scenario_bar = dbc.Row(
            [
                dcc.Store(id='reference_scenario_name_store'),
                dcc.Store(id='multi_scenario_names_store'),
                dbc.Col(
                    dcc.Dropdown(
                        id='top_menu_scenarios_drpdwn',
                        style={'width': '30vw'},
                    ),
                ),
                dbc.Col(
                    dbc.Button(
                        "Refresh", id='scenario_refresh_button', color="primary", className="ms-2", n_clicks=0
                    ),
                    width="auto",
                ),
            ],
            className="g-0 ms-auto flex-nowrap mt-3 mt-md-0",
            align="center",
        )

        navbar = dbc.Navbar(
            dbc.Container(
                [
                    html.A(
                        # Use row and col to control vertical alignment of logo / brand
                        dbc.Row(
                            [
                                dbc.Col(html.Img(src=self.get_logo_src(), height="50vh")),
                                # dbc.Col(dbc.NavbarBrand("Dashboard", className="ms-2")),
                                dbc.Col(dbc.NavbarBrand(self.get_navbar_brand_children(), className="ms-2")),
                            ],
                            align="center",
                            className="g-0",
                        ),
                        # href="/",
                        style={"textDecoration": "none"},
                    ),
                    # scenario_bar
                    dbc.NavbarToggler(id="navbar-toggler", n_clicks=0),  # The NavbarToggler is a hamburger menu that will only appear of the screen is very small
                    dbc.Collapse(
                        scenario_bar,
                        id="navbar-collapse",
                        is_open=False,
                        navbar=True,
                    ),
                ]
            ),
            color="dark",
            dark=True,
        )
        return navbar
    # def get_navbar(self):
    #     navbar = dbc.Navbar(
    #         dbc.Container(
    #             [
    #                 html.A(
    #                     # Use row and col to control vertical alignment of logo / brand
    #                     dbc.Row(
    #                         [
    #                             dbc.Col(html.Img(src=self.get_logo_src(), height="50vh")),
    #                             dbc.Col(dbc.NavbarBrand("Dashboard", className="ms-2")),
    #                         ],
    #                         # align="left",
    #                         className="g-0",
    #                     ),
    #                     style={"textDecoration": "none"},
    #                 ),
    #                 dcc.Dropdown(
    #                     id='top_menu_scenarios_drpdwn',
    #                     style={'width': '20vw'},
    #
    #                 ),
    #
    #                 html.Button('Refresh', id='scenario_refresh_button', n_clicks=0,
    #                             style={'width': '300px'}
    #                             ),
    #             ]
    #         ),
    #         color="dark",
    #         dark=True,
    #         fixed='top'
    #     )
    #     return navbar

    def get_sidebar(self):
        sidebar_style = {
            "position": "fixed",
            "top": self.header_height,
            "left": 0,
            "bottom": self.footer_height,
            "width": self.sidebar_width,
            "padding": "1rem 1rem",
        }
        sidebar = html.Div([
            html.Hr(),
            dbc.Nav([

                dbc.NavLink(
                    href=self.app.get_relative_path('/'),
                    children='Home'
                ),
                # dbc.NavLink(
                #     href=self.app.get_relative_path('/prepare-data'),
                #     children='Prepare Data'
                # ),
                # dbc.NavLink(
                #     href=self.app.get_relative_path('/run-model'),
                #     children='Run Model'
                # ),
                # dbc.NavLink(
                #     href=self.app.get_relative_path('/explore-solution'),
                #     children='Explore Solution'
                # ),
                # dbc.NavLink(
                #     href=app.get_relative_path('/visualization'),
                #     children='Visualization'
                # ),

                # dbc.Button(
                #     "Visualization",
                #     outline=True,
                #     id = 'collapse-button',
                #     className='mb-3',
                #     n_clicks=0,
                #     href=self.app.get_relative_path('/visualization'),
                #     # active=True,
                #     color='primary',
                #
                # ),

                # dbc.Collapse(
                #     children = visualization_menu_children,
                #     id = "collapse",
                #     is_open=True,
                #     navbar=True,
                # ),

            ], vertical=True, pills=True,)], style = sidebar_style
        )
        return sidebar

    def get_app_stores_div(self) -> html.Div:
        """A Div container with all globally (and thus permanently available) dcc.Stores in the app."""
        stores = html.Div(
            id = 'global_stores',  # Not necessary
            children = self.get_app_stores()
        )
        return stores

    def get_app_stores(self) -> List[dcc.Store]:
        """To be overridden"""
        stores = []
        return stores

    def set_cache_callbacks(self):
        pass

    def set_dash_callbacks(self):
        app = self.app

        @app.callback(
            Output('content', 'children'),
            [Input('url', 'pathname'), Input('top_menu_scenarios_drpdwn', 'value')],
            [State('reference_scenario_name_store', 'data'),
             State('multi_scenario_names_store', 'data')
             ]
        )
        def display_content(pathname, scenario_name, reference_scenario_name, multi_scenario_names):
            """High level call back to update the content section of the app.
            Triggers when either the URL or the scenario-dropdown changes."""
            # reference_scenario_name = None,  #TODO
            # multi_scenario_names = None  #TODO

            # print(f"Reference scenario = {reference_scenario_name}")
            # print(f"Multi scenario names = {multi_scenario_names}")
            print(f"Showing URL = {pathname}")
            return self.display_content_callback(pathname, scenario_name, reference_scenario_name, multi_scenario_names)

        @app.callback(
            [Output('top_menu_scenarios_drpdwn', 'options'), Output('top_menu_scenarios_drpdwn', 'value')],
            [Input('scenario_refresh_button', 'n_clicks')],
            [State('top_menu_scenarios_drpdwn', 'value')],
            # Read the state of scenarios dropdown. Do NOT define this as an Input, otherwise the refresh will be triggered when you change the scenario.
            # prevent_initial_call = True  # If True, would prevent clearing the cache. However, it will also prevent the initialization of the scenario drpdwn menu!
        )
        def refresh_scenarios(n_clicks, current_scenario_name:str):
            """Reads the currently selected scenario-name. Reloads the scenario names from DB.
            If the currently selected scenario is valid, this will be maintained as the selected value.
            Main callback from scenario dropdown (?)
            """
            options, initial_scenario_name = self.refresh_scenarios_dash_callback(current_scenario_name)
            return [options, initial_scenario_name]

        @app.callback(
            Output("navbar-collapse", "is_open"),
            [Input("navbar-toggler", "n_clicks")],
            [State("navbar-collapse", "is_open")],
        )
        def toggle_navbar_collapse(n, is_open):
            """Callback for NavbarToggler in top-menubar.
            The NavbarToggler is a hamburger menu that will only appear of the screen is very small.
            So in normal desktop use, this toggler and callback are irrelevant."""
            if n:
                return not is_open
            return is_open

    def shutdown(self):
        """Shuts-down the Flash web-server, releasing the port.
        Relevant in CPD, so the port gets released immediately.
        """
        from flask import request
        func = request.environ.get('werkzeug.server.shutdown')
        if func is None:
            raise RuntimeError('Not running with the Werkzeug Server')
        func()

    def get_table_schema(self, table_name) -> Optional[ScenarioTableSchema]:
        pass



# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional, List

from dash.exceptions import PreventUpdate

from dse_do_dashboard.main_pages.main_page import MainPage
from dash.dependencies import Input, Output
from dash import dcc, html
import dash_bootstrap_components as dbc

from dse_do_dashboard.utils.dash_common_utils import get_data_table_card_children, get_pivot_table_card_children


class ExploreSolutionPage(MainPage):
    def __init__(self, dash_app):
        super().__init__(dash_app,
                         page_name='Explore Solution',
                         page_id='explore-solution',
                         url='explore-solution',
                         )

    def get_layout(self, scenario_name: str = None, reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        output_tables = self.dash_app.get_output_table_names()
        layout = html.Div([

            dbc.Card([
                dbc.CardHeader('Output Table', style= {'fullscreen': True}),
                dbc.CardBody(
                    dcc.Dropdown(id='output_table_drpdwn',
                                 options=[ {'label': i, 'value': i}
                                           for i in output_tables],
                                 value=(output_tables[0] if len(output_tables) > 0 else None),  # In case there are no output tables
                                 style = {'width': '75vw','height': '2vw'},
                                 ),
                ),
            ], style = {'width': '80vw'}),

            dbc.Card([
                # dbc.CardHeader('Input Table'),
                dbc.CardBody(id = 'output_data_table_card',style = {'width': '79vw'} ),
                html.Div(id="output_data_table_div")
            ], style = {'width': '80vw'}),

            dbc.Card([

                dbc.CardBody(id = 'output_pivot_table_card',style = {'width': '79vw'}),
                html.Div(id="output_pivot_table_div"),
            ], style = {'width': '80vw'})

        ])
        return layout


    def update_data_and_pivot_output_table_callback(self, scenario_name, table_name: Optional[str]):
        """Body for the Dash callback.

        :param table_name (str): name of selected table. Can be None if no tables.

        Usage::

            @app.callback([Output('output_data_table_card', 'children'),
               # Output('output_pivot_table_div', 'children'),
               Output('output_pivot_table_card', 'children'),
               ],
              [Input('top_menu_scenarios_drpdwn', 'value'),
               Input('output_table_drpdwn', 'value')])
            def update_data_and_pivot_output_table(scenario_name, table_name):
                data_table_children, pivot_table_children = DA.update_data_and_pivot_output_table_callback(scenario_name, table_name)
                return [data_table_children, pivot_table_children]

        """
        if table_name is None:
            # In case there are no output tables
            raise PreventUpdate
        output_table_names = [table_name]
        pm = self.dash_app.get_plotly_manager(scenario_name, [], output_table_names)
        dm = pm.dm
        df = dm.outputs[table_name]
        table_schema = self.dash_app.get_table_schema(table_name)
        pivot_table_config = self.dash_app.get_pivot_table_config(table_name)
        data_table_children = get_data_table_card_children(df, table_name, table_schema)
        pivot_table_children = get_pivot_table_card_children(df, scenario_name, table_name, pivot_table_config)
        return data_table_children, pivot_table_children


    def set_dash_callbacks(self):
        """Define Dash callbacks for this page

        Will be called to register any callbacks
        :return:
        """
        app = self.dash_app.app

        @app.callback([Output('output_data_table_card', 'children'),
                       Output('output_pivot_table_card', 'children'),
                       ],
                      [Input('top_menu_scenarios_drpdwn', 'value'),
                       Input('output_table_drpdwn', 'value')])
        def update_data_and_pivot_output_table(scenario_name, table_name):
            data_table_children, pivot_table_children = self.update_data_and_pivot_output_table_callback(scenario_name, table_name)
            return [data_table_children, pivot_table_children]


# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

name = "dse_do_dashboard"

from .version import __version__
from .do_dash_app import DoDashApp
from .visualization_pages.visualization_page import VisualizationPage
from .visualization_pages.plotly_1_column_visualization_page import Plotly1ColumnVisualizationPage


import time
from typing import List, Dict

import pandas as pd
import dash
from dash import dcc, html, Output, Input, State, dash_table
import dash_bootstrap_components as dbc
from dash.exceptions import PreventUpdate


from dse_do_dashboard.main_pages.main_page import MainPage
from dse_do_dashboard.utils.domodelrunner import DoModelRunnerConfig


class RunModelPage(MainPage):
    """
    Challenge with long_callback:
    If you navigate away from this page while the long_callback is running, Dash will generate errors
    stating that input/output components do not exist:
    All components ('ids') that are used for inputs, outputs, states and progress (output) must continue to exists
    while the callback is running.
    Work-around:
    Use dcc.Store(s) on a global level:
    * lrc_job_trigger_store: the result of pressing the run button, along with the scenario_name and model_class
    * lrc_job_log_store: the log, i.e. the output of the long_callback
    * lrc_job_queue_data_store: the progress output of the long_callback
    Next, we have:
    * A button and a regular callback that store the trigger data (i.e. the user pressing the run button)
    * A regular callback watching the log store and update the log text area
    * A regular callback watching the queue data store and update the lrc job queue datatable

    """
    def __init__(self, dash_app,
                 page_name: str = 'Run Model',
                 page_id : str = 'run-model',
                 url: str = 'run-model',

                 ):
        super().__init__(dash_app,
                         page_name=page_name,
                         page_id=page_id,
                         url=url,
                         )
        self.runner_config_dict: Dict[str, DoModelRunnerConfig] = {config.runner_id: config for config in dash_app.get_do_model_runner_configs()}

    def get_layout(self, scenario_name: str = None, reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        configs = list(self.runner_config_dict.values())

        layout = html.Div([
            # dcc.Store(id='job_queue_data_store'),
            dcc.Dropdown(
                id='do_model_class_drpdwn',
                options=[{'label': 'None', 'value': 'None'}]+[
                    {'label': config.runner_name, 'value': config.runner_id}
                    for config in configs
                ],
                # value=(configs[0].runner_id if len(configs) > 0 else None),
                value=('None'),
                style = {'width': '28vw',
                         # 'background-color': '#FF4136'
                         }),
            dbc.Button('Run Model inline', id='run_model_inline', n_clicks=0, color="primary", className="me-1"),
            dbc.Button('Run Model LRC', id='run_model_lrc', n_clicks=0, color="primary", className="me-1", disabled=(not self.dash_app.enable_long_running_callbacks)),
            dbc.Button('Update JobQueue', id='refresh_queue_inline', n_clicks=0, color="primary", className="me-1"),

            # html.H1("------------------------"),
            dbc.Card([
                dbc.CardHeader('Job Queue (in-line)'),
                dbc.CardBody(
                    children=[
                        self.get_job_queue_data_table(id = 'job_queue_table_inline'),
                    ]
                ),
            ]),
            dbc.Card([
                dbc.CardHeader('Job Queue (long-running)'),
                dbc.CardBody(
                    children=[
                        self.get_job_queue_data_table(id = 'job_queue_table_lrc'),
                    ]
                ),
            ]),
            dbc.Card([
                dbc.CardHeader('Job Log (in-line)'),
                dbc.CardBody(
                    children=[
                        dcc.Textarea(id='log_inline',
                                     readOnly =True,
                                     style={'width': '100%', 'height': 200},
                                     ),
                    ]
                ),
            ]),
            dbc.Card([
                dbc.CardHeader('Job Log (long-running)'),
                dbc.CardBody(
                    children=[
                        dcc.Textarea(id='log_lrc',
                                     readOnly =True,
                                     style={'width': '100%', 'height': 200},
                                     ),
                    ]
                ),
            ]),
        ],
        )
        return layout

    def set_dash_callbacks(self):
        """Define Dash callbacks for this page

        Will be called to register any callbacks
        :return:
        """
        app = self.dash_app.app

        # Need to get the runner_config_dict as a 'global' variable so that the long_callback can be pickled
        # runner_config_dict = self.runner_config_dict

        def make_default_progress_data():
            """We do not have to specify columns"""
            df = pd.DataFrame()
            return df.to_dict('records')

        @app.callback(
            Output('log_inline', 'value'),
                           [Input('run_model_inline', 'n_clicks')],
                           [State('top_menu_scenarios_drpdwn', 'value'),
                            State('do_model_class_drpdwn', 'value'),
                            ],
                           prevent_initial_call=True,
                           )
        def run_model_inline_callback(
                n_clicks,
                scenario_name: str,
                do_model_class_name: str,
                ):

            # print("RunModelPage2.run_model_callback_inline")

            if do_model_class_name == 'None':
                raise PreventUpdate

            # time.sleep(0.2)

            runner_class = self.runner_config_dict[do_model_class_name].runner_class
            print(f"Runner class = {runner_class}")
            runner = runner_class(scenario_name, dash_app=self.dash_app)
            # runner = FruitClassRunner(scenario_name)
            runner.run()
            # time.sleep(4)

            log = f"Run {do_model_class_name} with scenario {scenario_name}\n" \
                  "Log: \n" \
                  f"{runner.log}"
            return log


        @app.callback(Output('job_queue_table_lrc', 'data'),
                      [Input('lrc_job_queue_data_store', 'data')],
                      )
        def refresh_queue_callback(data):
            """The idea is to place the Store in the top-menu bar and this is always there.
            That may prevent issues with the long_callback.
            However, it seems to be also sensitive towards the input button.
            So this work-around may not apply anyway.
            Not in use currently. Here as a backup if we need it"""
            # print("refresh_queue_callback")
            if data is None or len(data) == 0:
                raise PreventUpdate
            # print(f"refresh_queue_callback with data {data}")
            return data

        @app.callback(Output('job_queue_table_inline', 'data'),
                      Input('refresh_queue_inline', 'n_clicks'),
                      )
        def refresh_job_queue_inline(n_clicks):
            return self.get_job_queue_data()


        @app.callback(Output('lrc_job_trigger_store', 'data'),
                      Input('run_model_lrc', 'n_clicks'),
                      [State('top_menu_scenarios_drpdwn', 'value'),
                      State('do_model_class_drpdwn', 'value'),],
                      prevent_initial_call=True,
                      )
        def lrc_job_trigger(n_clicks,
                            scenario_name: str,
                            do_model_class_name: str,):
            # TODO: as an extra precaution use the ctx to check it was the button that triggered the callback?
            # Note that switching scenarios will trigger this callback. Despite the prevent_initial_call
            # However, since switching scenarios will regenerate this page, the n_clicks will be 0
            trigger = {'n_clicks':n_clicks, 'scenario' : scenario_name, 'do_model_class_name': do_model_class_name}
            # print(f"Initial Trigger = {trigger}")
            if n_clicks == 0 or do_model_class_name == 'None':
                raise PreventUpdate
            trigger = {'n_clicks':n_clicks, 'scenario_name' : scenario_name, 'do_model_class_name': do_model_class_name}
            # print(f"Store Trigger = {trigger}")
            return trigger

        @app.callback(Output('log_lrc', 'value'),
                      Input('lrc_job_log_store', 'data'),
                      # prevent_initial_call=True,
                      )
        def lrc_job_log_update(data):
            # print(f"lrc_job_log_update = {data}")
            if data is None:
                raise PreventUpdate
            return data

    def get_job_queue_data(self) -> List[Dict]:
        """Return a List[Dict] with job DoModelRunner queue data.
        As in df.to_dict('records')
        """
        rows = []
        for job in self.dash_app.job_queue:
            row = {'id' : job.id,
                   'model': job.__class__.__name__,
                   'scenario' : job.scenario_name,
                   'run_status' : job.run_status}
            rows.append(row)
        # if len(rows) > 0:
        #     df = pd.DataFrame(rows)
        # else:
        #     df = pd.DataFrame(columns=['id', 'runner_class', 'scenario_name', 'run_status'])
        return rows

    def get_job_queue_data_table(self, id: str) -> dash_table.DataTable:
        return dash_table.DataTable(
            id=id,
            columns=[{'name': 'scenario', 'id': 'scenario'},
                     {'name': 'model', 'id': 'model'},
                     {'name': 'run_status', 'id': 'run_status'}
                     ],
            fixed_rows={'headers': True},
            editable=True,
            # fixed_columns={'headers': False, 'data': 0}, # Does NOT create a horizontal scroll bar
            filter_action="native",
            sort_action="native",
            sort_mode="multi",
            style_cell={
                'textOverflow': 'ellipsis',  # See https://dash.plotly.com/datatable/width to control column-name width
                'maxWidth': 0,               # Needs to be here for the 'ellipsis' option to work
                'overflow' : 'hidden',
                'font_family': 'sans-serif',
                'font_size': '12px',
                'textAlign': 'left'},
            style_table={
                'maxHeight': '400px',
                'overflowY': 'scroll'
            },
        )


# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from dse_do_dashboard.main_pages.main_page import MainPage
from dash import dcc, html


class NotFoundPage(MainPage):
    def __init__(self, dash_app):
        super().__init__(dash_app,
                         page_name='Not Found',
                         page_id='not-found',
                         url='not-found',
                         )

    def get_layout(self, error_message: str = ""):
        layout = html.Div([

            html.H1("Error 404 - Page not found"),
            dcc.Textarea(
                id='not_found',
                # value='Page not found',
                value=error_message,
                style={'width': '100%', 'height': 300},
            ),

        ])
        return layout


# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from abc import abstractmethod
from typing import List, NamedTuple

from dash import dcc
import dash_bootstrap_components as dbc
import plotly.graph_objs as go

from dse_do_dashboard.do_dash_app import DoDashApp
from dse_do_dashboard.visualization_pages.visualization_page import VisualizationPage
from dse_do_utils.plotlymanager import PlotlyManager


class PlotlyRowsVisualizationPage(VisualizationPage):
    """Abstract class. Creates a layout of Plotly Figures in a row format.
    Each row can contain a variable set of figures.
    Override the method `get_plotly_figures`.
    """
    def __init__(self, dash_app: DoDashApp, page_name: str = 'Default', page_id: str = 'default', url: str = 'default',
                 input_table_names: List[str] = [], output_table_names: List[str] = []):
        super().__init__(dash_app=dash_app,
                         page_name=page_name,
                         page_id=page_id,
                         url=url,
                         input_table_names=input_table_names,
                         output_table_names=output_table_names,
                         )

    @abstractmethod
    def get_plotly_figures(self, pm: PlotlyManager) -> List[List[go.Figure]]:
        """
        :returns List of Lists of Plotly Figures. One for each row.
        """
        rows=[]
        return rows

    def get_layout_children(self, pm: PlotlyManager):
        """Creates a layout for KPI gauges with multiple rows and columns.
        Each row can contain any number of columns."""
        figures = self.get_plotly_figures(pm=pm)

        figure_style = {'height': '20vh', 'width': '20vw', 'margin-left': 'auto', 'margin-right': 'auto',
                       'display': 'block'}
        layout_childen = []

        for figure_row in figures:
            row_layout = dbc.Row([
                dbc.Col(
                    dbc.Card([
                        dbc.CardBody(
                            dcc.Graph(
                                style=figure_style,
                                figure=figure
                            )
                        )
                    ])
                )
                for figure in figure_row
            ])
            layout_childen.append(row_layout)

        return layout_childen



# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from typing import List
from dash import html
# from dse_do_dashboard.do_dash_app import DoDashApp
from dse_do_dashboard.main_pages.not_found import NotFoundPage
from dse_do_utils.plotlymanager import PlotlyManager


class VisualizationPage(ABC):
    """Abstract class

    Subclass should override/implement:
    * get_layout_children(self)
    """
    def __init__(self, dash_app, page_name:str='Default', page_id:str='default', url:str='default',
                 input_table_names: List[str] = [], output_table_names: List[str] = [],
                 enable_reference_scenario: bool = False,
                 enable_multi_scenario: bool = False):
        """
        Best practice to specify the minimal set of input and output tables.
        Loading less tables speeds of the response time of the dashboard.
        Shortcut to load all (input or output) tables: ['*']

        Args:
            dash_app (DoDashApp)
            page_name (str): Name of page in UI
            page_id (str): Internal unique ID
            url (str): url name
            input_table_names (List[str]): list of input table names required loading by the combined visualizations on this page. Or ['*'] for all input tables.
            output_table_names (List[str]): list of output table names required loading by the combined visualizations on this page. Or ['*'] for all output tables.
            enable_reference_scenario (bool): if True and if reference_scenario_name not None, will load the reference scenario and add a ref_dm dataManager to the PlotlyManager
            enable_multi_scenario (bool): if True and if there are more than 1 scenarios to compare in multi_scenario_names, will load multiple scenarios (including the currently selected) in ms_inputs and ms_outputs of the PlotlyManager
        TODO:
        * dash_app: DoDashApp. Due to circular imports, cannot specify type!
        """
        self.dash_app = dash_app
        self.page_name = page_name  # As shown in UI
        self.page_id = page_id  # Used for Dash pattern-matching callback
        self.url = url  # URL, i.e. web address: `host:port/visualization/url`

        self.input_table_names = input_table_names  # Names of input tables required for any of the plots in this page
        self.output_table_names = output_table_names  # Names of output tables required for any of the plots in this page
        self.enable_reference_scenario = enable_reference_scenario
        self.enable_multi_scenario = enable_multi_scenario
        # if len(input_table_names) == 1 and input_table_names[0] == '*':
        #     self.input_table_names = dash_app.get_input_table_names()
        # else:
        #     self.input_table_names = input_table_names  # Names of input tables required for any of the plots in this page

    def get_layout(self, scenario_name: str,
                   reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        """Returns layout of visualization page.
        TODO:
        * We are NOT using pattern-matching callback anynore, so the `id` of the Div is no long relevant
        * Do we need 2 Divs? (This is what worked so far)
        """
        self.pm = self.get_plotly_manager(scenario_name,
                                          reference_scenario_name=reference_scenario_name,
                                          multi_scenario_names=multi_scenario_names,
                                          enable_reference_scenario=self.enable_reference_scenario,
                                          enable_multi_scenario=self.enable_multi_scenario)
        layout = html.Div([
            html.Div(
                id={
                    'type': 'tab_layout',
                    'index': self.page_id
                },
                children=self.get_layout_children(self.pm),
                # children=self.get_layout_children(self.get_plotly_manager(scenario_name,
                #                                                           reference_scenario_name=reference_scenario_name,
                #                                                           multi_scenario_names=multi_scenario_names,
                #                                                           enable_reference_scenario=self.enable_reference_scenario,
                #                                                           enable_multi_scenario=self.enable_multi_scenario))
            )
        ])
        return layout

    @abstractmethod
    def get_layout_children(self, pm: PlotlyManager) -> List:
        """Abstract method. To be overridden.
        Create a list of components that will be added as children of a Div.
        Use dbc.Row and dbc.Column to create a layout.
        Embed each Plotly figure in a dcc.Graph, within a dbc.Card.

        Args:
            pm (PlotlyManager): PlotlyManager subclass instantiated with the specified in- and output tables for this VisualizationPage

        Typical usage::

            def get_layout_children(self, pm: PlotlyManager):
                layout_children = [

                    dbc.Row(
                        dbc.Col(
                            dbc.Card([
                                dbc.CardBody(
                                    dcc.Graph(
                                        figure=pm.create_my_plotly_figure(),
                                    )
                                )
                            ])
                        )
                    ),
                ]
                return layout_children

        """
        layout = NotFoundPage().get_layout("Default Visualization Page. Need to override the `get_layout_children` method.")
        return layout

    def get_input_table_names(self) -> List[str]:
        """Return the specified table-names.
        Does the 'pattern-matching' based on the '*' option to include all tables.
        Can only be done dynamically (i.e. not in the constructor), because the self.dash_app.dbm might not have been initialized properly
        :return:
        """
        if len(self.input_table_names) == 1 and self.input_table_names[0] == '*':
            input_table_names = self.dash_app.get_input_table_names()
        else:
            input_table_names = self.input_table_names
        return input_table_names

    def get_output_table_names(self) -> List[str]:
        """Return the specified table-names.
        Does the 'pattern-matching' based on the '*' option to include all tables.
        :return:
        """
        if len(self.output_table_names) == 1 and self.output_table_names[0] == '*':
            output_table_names = self.dash_app.get_output_table_names()
        else:
            output_table_names = self.output_table_names
        return output_table_names

    def get_plotly_manager(self, scenario_name: str, reference_scenario_name: str = None, multi_scenario_names: List[str] = None,
                           enable_reference_scenario: bool = False,
                           enable_multi_scenario: bool = False):
        # reference_scenario_name = None
        # multi_scenario_names = None
        return self.dash_app.get_plotly_manager(scenario_name,
                                                self.get_input_table_names(),
                                                self.get_output_table_names(),
                                                reference_scenario_name=reference_scenario_name,
                                                multi_scenario_names=multi_scenario_names,
                                                enable_reference_scenario=enable_reference_scenario,
                                                enable_multi_scenario=enable_multi_scenario
                                                )

    def set_dash_callbacks(self):
        """Define Dash callbacks for this page.
        Will be called to register any callbacks.

        For the annotation `@app.callback`, get `app` from `app = self.dash_app.app`.

        :return:

        Usage::

            def set_dash_callbacks(self):
                app = self.dash_app.app

                @app.callback(Output('container_button_basic', 'children'),
                              [Input('run_model', 'n_clicks')])
                def run_model_callback(n_clicks):
                    return self.run_model_callback(n_clicks)

        """
        app = self.dash_app.app
        pass


from abc import abstractmethod
from typing import List
from dash import html

from dash import dcc
import dash_bootstrap_components as dbc
from plotly.graph_objs import Figure

# from dse_do_dashboard.do_dash_app import DoDashApp
from dse_do_dashboard.visualization_pages.visualization_page import VisualizationPage
# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from dse_do_utils.plotlymanager import PlotlyManager


class FoliumColumnVisualizationPage(VisualizationPage):
    """Shows one or more folium maps in single column."""
    def __init__(self, dash_app, page_name:str='Default', page_id:str='default', url:str='default',
                 input_table_names: List[str] = [], output_table_names: List[str] = [],
                 enable_reference_scenario: bool = False,
                 enable_multi_scenario: bool = False):
        super().__init__(dash_app=dash_app,
                         page_name=page_name,
                         page_id=page_id,
                         url=url,
                         input_table_names = input_table_names,
                         output_table_names = output_table_names,
                         enable_reference_scenario=enable_reference_scenario,
                         enable_multi_scenario=enable_multi_scenario
                         )

    @abstractmethod
    def get_folium_maps(self, pm: PlotlyManager) -> List:
        """
        :returns List[folium.Map]
        """
        return []

    def get_layout_children(self, pm: PlotlyManager):
        """Automatic 1 Column layout of all Plotly Figures.
        Vertical height is driven from the height defined in the Plotly figure.

        :param pm:
        :return:
        """
        layout_children = [
            dbc.Row(
                dbc.Col(
                    dbc.Card([
                        ##       dbc.CardHeader(""),
                        dbc.CardBody(
                            html.Iframe(id='folium-map',  width='100%', height=800,
                                        srcDoc = m.get_root().render()
                                        )
                        )
                    ])
                    # , width=12
                )
            )
            for m in self.get_folium_maps(pm)
        ]

        return layout_children



from abc import ABC, abstractmethod
from dse_do_utils.scenariodbmanager import Inputs, Outputs, ScenarioDbManager


class DoModelRunner(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def run(self, inputs: Inputs) -> Outputs:
        """Abstract method"""
        return {}


class DoNotebookRunner(DoModelRunner):
    def __init__(self, notebook_filepath: str):
        super().__init__()
        self.notebook_filepath = notebook_filepath

    def run(self, inputs: Inputs) -> Outputs:
        """TODO"""
        return {}


class DoDeployRunner(DoModelRunner):
    def __init__(self):
        super().__init__()

    def run(self, inputs: Inputs) -> Outputs:
        """TODO"""
        return {}


class DoDbJobRunner():
    """Runs a DO Model based on a scenario in the DB."""
    def __init__(self, dbm: ScenarioDbManager, scenario_name: str, model: DoModelRunner):
        self.dbm = dbm
        self.scenario_name = scenario_name
        self.model = model

    def run(self):
        inputs = self.dbm.read_scenario_input_tables_from_db(self.scenario_name)
        outputs = self.model.run(inputs)
        self.dbm.update_scenario_output_tables_in_db(scenario_name=self.scenario_name, outputs=outputs)


from abc import ABC, abstractmethod

# from dse_do_dashboard import DoDashApp
from dse_do_utils.deployeddomodel import DeployedDOModel
from dse_do_utils.scenariodbmanager import Inputs, Outputs, ScenarioDbManager
from dse_do_dashboard.utils.donotebookrunner import DoNotebookRunner
from typing import NamedTuple
from typing import Type


class DoModelRunner(ABC):
    """
    TODO: For all model classes, after completion grab run-time, log, solve_status
    """
    def __init__(self, scenario_name: str,
                 dash_app = None,  # DoDashApp, cannot declare due to circular import
                 database_manager_class = None,
                 db_credentials = None,
                 schema: str = None,
                 db_echo: bool = False,
                 ):
        # self.dash_app = dash_app
        self.dash_app = dash_app
        self.database_manager_class = database_manager_class
        self.db_credentials = db_credentials
        self.schema = schema
        self.db_echo = db_echo
        if self.dash_app is None:
            self.dbm = self.create_database_manager_instance()
        else:
            self.dbm = self.dash_app.dbm  # Re-use existing dbm if available:
            self.dash_app.job_queue.append(self)  # TODO: hack to do a job-queue. Be careful with global variables in Dash!
        self.scenario_name = scenario_name
        self.log = ""
        self.code = ""  # Only applies to DoNotebookRunner
        self.run_status: str = "Initializing"
        self.id = 'my id'  # TODO: generate unique ID

    def create_database_manager_instance(self) -> ScenarioDbManager:
        """Create an instance of a ScenarioDbManager.
        The default implementation uses the database_manager_class from the constructor.
        Optionally, override this method."""
        if self.database_manager_class is not None and self.db_credentials is not None:
            print(f"Connecting to DB2 at {self.db_credentials['host']}, schema = {self.schema}")
            dbm = self.database_manager_class(credentials=self.db_credentials, schema=self.schema, echo=self.db_echo)
        else:
            print("Error: either specifiy `database_manager_class`, `db_credentials` and `schema`, or override `create_database_manager_instance`.")
        return dbm

    def run(self):
        """
        1. Load inputs from DB
        2. Run model
        3. Write outputs back to DB
        """
        self.run_status = "Loading input data"
        inputs = self.load_inputs()
        self.run_status = "Running Model"
        outputs = self.run_model(inputs)
        self.run_status = "Saving output data"
        self.update_outputs(outputs)
        self.run_status = "Done"

    def load_inputs(self) -> Inputs:
        return self.dbm.read_scenario_input_tables_from_db(self.scenario_name)

    @abstractmethod
    def run_model(self, inputs: Inputs) -> Outputs:
        """Abstract method"""
        return {}

    def update_outputs(self, outputs: Outputs):
        pass
        # self.dbm.update_scenario_output_tables_in_db(scenario_name=self.scenario_name, outputs=outputs)


class DoNotebookModelRunner(DoModelRunner):
    def __init__(self, scenario_name: str, notebook_filepath: str,
                 my_globals, my_locals,
                 dash_app = None,  # DoDashApp, cannot declare due to circular import
                 database_manager_class = None,
                 db_credentials = None,
                 schema: str = None,
                 db_echo: bool = False,
                 ):
        super().__init__(scenario_name,
                         dash_app=dash_app,
                         database_manager_class=database_manager_class,
                         db_credentials=db_credentials,
                         schema=schema,
                         db_echo=db_echo
                         )
        self.notebook_filepath = notebook_filepath
        self.globals = my_globals
        self.locals = my_locals

    def run_model(self, inputs: Inputs) -> Outputs:
        """TODO"""
        runner = DoNotebookRunner(self.notebook_filepath, self.globals, self.locals)
        outputs = runner.run(inputs)
        self.log = runner.log
        self.code = runner.code
        return outputs

class DoClassModelRunner(DoModelRunner):
    """
    TODO: grab the self.log so we can use this in the UI
    """
    def __init__(self, scenario_name: str, data_manager_class,
                 optimization_engine_class,
                 dash_app = None,  # DoDashApp, cannot declare due to circular import
                 database_manager_class = None,
                 db_credentials = None,
                 schema: str = None,
                 db_echo: bool = False,
                 ):
        super().__init__(scenario_name,
                         dash_app=dash_app,
                         database_manager_class=database_manager_class,
                         db_credentials=db_credentials,
                         schema=schema,
                         db_echo=db_echo
                         )
        self.data_manager_class = data_manager_class
        self.optimization_engine_class = optimization_engine_class

    def run_model(self, inputs: Inputs) -> Outputs:
        """Run an OptimizationEngine class"""
        dm = self.data_manager_class(inputs=inputs)
        engine = self.optimization_engine_class(data_manager=dm)
        outputs = engine.run()
        return outputs


class DoDeployedModelRunner(DoModelRunner):
    def __init__(self, scenario_name: str,
                 wml_credentials,
                 space_name: str, deployment_id: str,
                 dash_app = None,  # DoDashApp, cannot declare due to circular import
                 database_manager_class = None,
                 db_credentials = None,
                 schema: str = None,
                 db_echo: bool = False,
                 ):
        super().__init__(scenario_name=scenario_name,
                         dash_app=dash_app,
                         database_manager_class=database_manager_class,
                         db_credentials=db_credentials,
                         schema=schema,
                         db_echo=db_echo
                         )
        self.wml_credentials = wml_credentials
        self.space_name = space_name
        self.deployment_id = deployment_id

    def run_model(self, inputs: Inputs) -> Outputs:
        """"""
        mdl = DeployedDOModel(self.wml_credentials, space_name=self.space_name, deployment_id=self.deployment_id)
        job_details = mdl.solve(inputs)
        return mdl.outputs


# class DoDbJobRunner():
#     """Runs a DO Model based on a scenario in the DB."""
#     def __init__(self, dbm: ScenarioDbManager, scenario_name: str, model: DoModelRunner):
#         self.dbm = dbm
#         self.scenario_name = scenario_name
#         self.model = model
#
#     def run(self):
#         inputs = self.dbm.read_scenario_input_tables_from_db(self.scenario_name)
#         outputs = self.model.run(inputs)
#         self.dbm.update_scenario_output_tables_in_db(scenario_name=self.scenario_name, outputs=outputs)


class DoModelRunnerConfig(NamedTuple):
    runner_name: str
    runner_id: str
    runner_class: Type[DoModelRunner]


# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Source of truth for the dse_do_utils version.
The versions in `setup.py` and `/docs/source/conf.py` are automatically populated from here.

Best practice to keep version here, in a separate file.
See https://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
"""

__version__ = "0.1.2.1b0"



# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Optional, List

import pandas as pd
from dash.exceptions import PreventUpdate

from dse_do_dashboard.main_pages.main_page import MainPage
from dash.dependencies import Input, Output, State
from dash import dcc, html
import dash_bootstrap_components as dbc
import pprint

from dse_do_dashboard.utils.dash_common_utils import get_data_table_card_children, get_pivot_table_card_children


class PrepareDataPage(MainPage):
    def __init__(self, dash_app,
                 page_name: str = 'Prepare Data',
                 page_id: str = 'prepare-data',
                 url: str = 'prepare-data'):
        super().__init__(dash_app,
                         page_name=page_name,
                         page_id=page_id,
                         url=url,
                         )

    def get_layout(self, scenario_name: str = None, reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        # input_tables = self.dash_app.get_input_table_names()
        layout = html.Div([
            self.get_input_table_selection_card(),
            # dbc.Card([
            #     dbc.CardHeader('Input Table', style= {'fullscreen':True}),
            #     dbc.CardBody(
            #         dcc.Dropdown(id='input_table_drpdwn',
            #                      options=[ {'label': i, 'value': i}
            #                                for i in input_tables],
            #                      value=input_tables[0],
            #                      style = {'width': '75vw','height':'2vw'},
            #                      ),
            #     ),
            # ], style = {'width': '80vw'}),

            dbc.Card([
                # dbc.CardHeader('Input Table'),
                dbc.CardBody(id='input_data_table_card', style={'width': '79vw'} ),
                html.Div(id="input_data_table_div"),  # TODO: not being used? If so cleanup.
            ], style={'width': '80vw'}),

            # dbc.Card([
            #
            #     dbc.CardBody(id = 'input_pivot_table_card',style = {'width': '79vw'}),
            #     html.Div(id="input_pivot_table_div"),
            # ], style = {'width': '80vw'})
            self.get_pivot_table_card(),

        ])
        return layout

    def get_input_table_selection_card(self):
        input_tables = self.dash_app.get_input_table_names()
        card = dbc.Card([
            dbc.CardHeader('Input Table', style={'fullscreen': True}),
            dbc.CardBody(
                dcc.Dropdown(id='input_table_drpdwn',
                             options=[ {'label': i, 'value': i}
                                       for i in input_tables],
                             value=(input_tables[0] if len(input_tables) > 0 else None),  # Handle zero input tables (rare)
                             style = {'width': '75vw', 'height': '2vw'},
                             ),
            ),
        ], style={'width': '80vw'})
        return card

    def get_pivot_table_card(self):
        card = dbc.Card([
            dbc.CardBody(id='input_pivot_table_card', style={'width': '79vw'}),
            html.Div(id="input_pivot_table_div"),
        ], style={'width': '80vw'})
        return card

    def update_data_and_pivot_input_table_callback(self, scenario_name: str, table_name: Optional[str]):
        """Body for the Dash callback.

        :param scenario_name: name of scenario
        :param table_name: name of input table. Theoretically can be None if no input tables

        Usage::

            @app.callback([Output('input_data_table_card', 'children'),
            Output('input_pivot_table_card', 'children')],
            [Input('top_menu_scenarios_drpdwn', 'value'),
            Input('input_table_drpdwn', 'value')])
            def update_data_and_pivot_input_table(scenario_name, table_name):
                data_table_children, pivot_table_children = DA.update_data_and_pivot_input_table_callback(scenario_name, table_name)
                return [data_table_children, pivot_table_children]

        """
        # print(f"update_data_and_pivot_input_table for {table_name} in {scenario_name}")
        if table_name is None:
            # In case there are no input tables (rare)
            raise PreventUpdate
        input_table_names = [table_name]
        pm = self.dash_app.get_plotly_manager(scenario_name, input_table_names, [])
        dm = pm.dm
        df = self.dash_app.get_table_by_name(dm=dm, table_name=table_name, index=False, expand=False)
        table_schema = self.dash_app.get_table_schema(table_name)
        pivot_table_config = self.dash_app.get_pivot_table_config(table_name)
        data_table_children = get_data_table_card_children(df, table_name, table_schema, editable=True)
        pivot_table_children = get_pivot_table_card_children(df, scenario_name, table_name, pivot_table_config)
        return data_table_children, pivot_table_children

    def set_dash_callbacks(self):
        """Define Dash callbacks for this page

        Will be called to register any callbacks
        :return:
        """
        app = self.dash_app.app

        @app.callback([Output('input_data_table_card', 'children'),
                       Output('input_pivot_table_card', 'children')],
                      [Input('top_menu_scenarios_drpdwn', 'value'),
                       Input('input_table_drpdwn', 'value')])
        def update_data_and_pivot_input_table(scenario_name:str, table_name:str):
            data_table_children, pivot_table_children = self.update_data_and_pivot_input_table_callback(scenario_name, table_name)
            return [data_table_children, pivot_table_children]




# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import time
from typing import List

from dse_do_dashboard.main_pages.main_page import MainPage
from dash import dcc, html, Output, Input, State
import dash_bootstrap_components as dbc

# import diskcache
# from dash.long_callback import DiskcacheLongCallbackManager
# cache = diskcache.Cache("./cache")
# long_callback_manager = DiskcacheLongCallbackManager(cache)

class RunModelPage_Old(MainPage):
    def __init__(self, dash_app):
        super().__init__(dash_app,
                         page_name='Run Model',
                         page_id='run-model',
                         url='run-model',
                         )

    def get_layout(self, scenario_name: str = None, reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        layout = html.Div([

            dbc.Button('Run Model', id='run_model', n_clicks=0, color="primary", className="me-1"),
            # html.Button('Run Model', id='run_model', n_clicks=0),

            html.Div(id='container_button_basic',
                     children='Enter a value and press submit'),

            # ddk.Block(
            #     width=30,
            #     children=ddk.ControlCard(
            #         width=100,
            #         children=[
            #             ddk.ControlItem(
            #                 # label='xxx',
            #                 children=dcc.Dropdown(
            #                     id='scenarios',
            #                     options=[
            #                         {'label': i, 'value': i}
            #                         for i in names
            #                     ],
            #                     # value=scenarios_df.scenarioName[0]
            #                 )
            #             ),
            #         ]
            #
            #     )
            # ),

        ])
        return layout

    # @app.callback(
    #         Output('container-button-basic', 'children'),
    #         [Input('run_model', 'n_clicks')])
    def run_model_callback(self, n_clicks, scenario_name):
        """Callback for pressing `run-model` button
        Usage:
        1. On index.py add::

            @app.callback(
                    Output('container-button-basic', 'children'),
                    [Input('run_model', 'n_clicks')])
            def run_model_callback(self, n_clicks):
                DA.run_model_callback(n_clicks)

        2. On DashApp add method

            def run_model_callback(self, n_clicks):
                self.get_run_model_page().run_model_callback(n_clicks)
        """
        # self.dash_app.tmp_run_notebook(scenario_name)  # HACK!!!!!!!
        return f"The button has been clicked {n_clicks} times for {scenario_name}"

    def set_dash_callbacks(self):
        """Define Dash callbacks for this page

        Will be called to register any callbacks
        :return:
        """
        app = self.dash_app.app

        @app.callback(Output('container_button_basic', 'children'),
                      [Input('run_model', 'n_clicks')],
                      State('top_menu_scenarios_drpdwn', 'value'),
                      manager=self.dash_app.long_callback_manager,
                      prevent_initial_call=True
        )
        def run_model_callback(n_clicks, scenario_name):
            # time.sleep(2.0)
            return self.run_model_callback(n_clicks, scenario_name)




# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from multiprocessing.pool import ThreadPool
from typing import Dict, List, NamedTuple, Any, Optional

import pandas as pd
import sqlalchemy
from dse_do_utils.scenariodbmanager import ScenarioDbManager, ScenarioDbTable, DbCellUpdate

#  Typing aliases
Inputs = Dict[str, pd.DataFrame]
Outputs = Dict[str, pd.DataFrame]

#########################################
# Added to dse-do-utils v 0.5.3.2b
#########################################
# class DbCellUpdate(NamedTuple):
#     scenario_name: str
#     table_name: str
#     row_index: List[Dict[str, Any]]  # e.g. [{'column': 'col1', 'value': 1}, {'column': 'col2', 'value': 'pear'}]
#     column_name: str
#     current_value: Any
#     previous_value: Any  # Not used for DB operation
#     row_idx: int  # Not used for DB operation


class ScenarioDbManagerUpdate(ScenarioDbManager):
    """
    DEPRECATED - was used to develop DB features that are now migrated to dse-do-utils
    To test editable tables. Supports DB updates of edits.
    Supports CRUD operations on scenarios.
    Changes are added to dse-do-utils v 0.5.4.0b
    """

    def __init__(self, input_db_tables: Dict[str, ScenarioDbTable], output_db_tables: Dict[str, ScenarioDbTable],
                 credentials=None, schema: str = None, echo: bool = False, multi_scenario: bool = True,
                 enable_transactions: bool = True, enable_sqlite_fk: bool = True):
        super().__init__(input_db_tables, output_db_tables, credentials, schema, echo, multi_scenario, enable_transactions, enable_sqlite_fk)


    ############################################################################################
    # Update scenario
    # Added to dse-do-utils v 0.5.3.2b
    ############################################################################################
    # def update_cell_changes_in_db(self, db_cell_updates: List[DbCellUpdate]):
    #     """Update a set of cells in the DB.
    #
    #     :param db_cell_updates:
    #     :return:
    #     """
    #     if self.enable_transactions:
    #         print("Update cells with transaction")
    #         with self.engine.begin() as connection:
    #             self._update_cell_changes_in_db(db_cell_updates, connection=connection)
    #     else:
    #         self._update_cell_changes_in_db(db_cell_updates, connection=self.engine)

    # def _update_cell_changes_in_db(self, db_cell_updates: List[DbCellUpdate], connection):
    #     """Update an ordered list of single value changes (cell) in the DB."""
    #     for db_cell_change in db_cell_updates:
    #         self._update_cell_change_in_db(db_cell_change, connection)

    # def _update_cell_change_in_db(self, db_cell_update: DbCellUpdate, connection):
    #     """Update a single value (cell) change in the DB."""
    #     # db_table_name = self.db_tables[db_cell_update.table_name].db_table_name
    #     # column_change = f"{db_cell_update.column_name} = '{db_cell_update.current_value}'"
    #     # scenario_condition = f"scenario_name = '{db_cell_update.scenario_name}'"
    #     # pk_conditions = ' AND '.join([f"{pk['column']} = '{pk['value']}'" for pk in db_cell_update.row_index])
    #     # old_sql = f"UPDATE {db_table_name} SET {column_change} WHERE {pk_conditions} AND {scenario_condition};"
    #
    #     db_table: ScenarioDbTable = self.db_tables[db_cell_update.table_name]
    #     t: sqlalchemy.Table = db_table.get_sa_table()
    #     pk_conditions = [(db_table.get_sa_column(pk['column']) == pk['value']) for pk in db_cell_update.row_index]
    #     target_col: sqlalchemy.Column = db_table.get_sa_column(db_cell_update.column_name)
    #     sql = t.update().where(sqlalchemy.and_((t.c.scenario_name == db_cell_update.scenario_name), *pk_conditions)).values({target_col:db_cell_update.current_value})
    #     # print(f"_update_cell_change_in_db = {sql}")
    #
    #     connection.execute(sql)

    ############################################################################################
    # Missing CRUD operations on scenarios in DB:
    # - Delete scenario
    # - Duplicate scenario
    # - Rename scenario
    ############################################################################################
    # Migrated 2022-01-10
    # def delete_scenario_from_db(self, scenario_name: str):
    #     """Delete a scenario. Uses a transaction (when enabled)."""
    #     if self.enable_transactions:
    #         print("Delete scenario within a transaction")
    #         with self.engine.begin() as connection:
    #             self._delete_scenario_from_db(scenario_name=scenario_name, connection=connection)
    #     else:
    #         self._delete_scenario_from_db(scenario_name=scenario_name, connection=self.engine)

    # def duplicate_scenario_in_db(self, source_scenario_name: str, target_scenario_name: str):
    #     """Duplicate a scenario. Uses a transaction (when enabled)."""
    #     if self.enable_transactions:
    #         print("Duplicate scenario within a transaction")
    #         with self.engine.begin() as connection:
    #             self._duplicate_scenario_in_db(connection, source_scenario_name, target_scenario_name)
    #     else:
    #         self._duplicate_scenario_in_db(self.engine, source_scenario_name, target_scenario_name)
    #
    # def _duplicate_scenario_in_db(self, connection, source_scenario_name: str, target_scenario_name: str = None):
    #     """Is fully done in DB using SQL in one SQL execute statement
    #     :param source_scenario_name:
    #     :param target_scenario_name:
    #     :param connection:
    #     :return:
    #     """
    #     if target_scenario_name is None:
    #         new_scenario_name = self._find_free_duplicate_scenario_name(source_scenario_name)
    #     elif self._check_free_scenario_name(target_scenario_name):
    #         new_scenario_name = target_scenario_name
    #     else:
    #         raise ValueError(f"Target name for duplicate scenario '{target_scenario_name}' already exists.")
    #
    #     # inputs, outputs = self.read_scenario_from_db(source_scenario_name)
    #     # self._replace_scenario_in_db_transaction(scenario_name=new_scenario_name, inputs=inputs, outputs=outputs,
    #     #                                          bulk=True, connection=connection)
    #     self._duplicate_scenario_in_db_sql(connection, source_scenario_name, new_scenario_name)
    #
    # def _duplicate_scenario_in_db_sql(self, connection, source_scenario_name: str, target_scenario_name: str = None):
    #     """
    #     :param source_scenario_name:
    #     :param target_scenario_name:
    #     :param connection:
    #     :return:
    #
    #     See https://stackoverflow.com/questions/9879830/select-modify-and-insert-into-the-same-table
    #
    #     Problem: the table Parameter/parameters has a column 'value' (lower-case).
    #     Almost all of the column names in the DFs are lower-case, as are the column names in the ScenarioDbTable.
    #     Typically, the DB schema converts that the upper-case column names in the DB.
    #     But probably because 'VALUE' is a reserved word, it does NOT do this for 'value'. But that means in order to refer to this column in SQL,
    #     one needs to put "value" between double quotes.
    #     Problem is that you CANNOT do that for other columns, since these are in upper-case in the DB.
    #     Note that the kpis table uses upper case 'VALUE' and that seems to work fine
    #
    #     Resolution: use SQLAlchemy to construct the SQL. Do NOT create SQL expressions by text manipulation.
    #     SQLAlchemy has the smarts to properly deal with these complex names.
    #     """
    #     if target_scenario_name is None:
    #         new_scenario_name = self._find_free_duplicate_scenario_name(source_scenario_name)
    #     elif self._check_free_scenario_name(target_scenario_name):
    #         new_scenario_name = target_scenario_name
    #     else:
    #         raise ValueError(f"Target name for duplicate scenario '{target_scenario_name}' already exists.")
    #
    #     batch_sql=False  # BEWARE: batch = True does NOT work!
    #     sql_statements = []
    #
    #     # 1. Insert scenario in scenario table
    #     # sql_insert = f"INSERT INTO SCENARIO (scenario_name) VALUES ('{new_scenario_name}')"  # Old
    #     # sa_scenario_table = list(self.input_db_tables.values())[0].table_metadata  # Scenario table must be the first
    #     sa_scenario_table = list(self.input_db_tables.values())[0].get_sa_table()  # Scenario table must be the first
    #     sql_insert = sa_scenario_table.insert().values(scenario_name = new_scenario_name)
    #     # print(f"_duplicate_scenario_in_db_sql - Insert SQL = {sql_insert}")
    #     if batch_sql:
    #         sql_statements.append(sql_insert)
    #     else:
    #         connection.execute(sql_insert)
    #
    #     # 2. Do 'insert into select' to duplicate rows in each table
    #     for scenario_table_name, db_table in self.db_tables.items():
    #         if scenario_table_name == 'Scenario':
    #             continue
    #         # else:
    #         #     table_column_names = db_table.get_df_column_names()
    #
    #         # print(f"#####TABLE METADATA: {type(db_table.table_metadata)}")
    #         # print(t1.insert().from_select(['a', 'b'], t2.select().where(t2.c.y == 5)))
    #
    #         # print(f"Columns for {scenario_table_name}: {table_column_names}")
    #         # target_column_names = table_column_names.copy()
    #         # # target_column_names = [f"{db_table.db_table_name}.{n}" for n in target_column_names]
    #         # target_columns_txt = ','.join(target_column_names)
    #         # # target_columns_txt = ','.join([f'"{n}"' for n in target_column_names])
    #         # source_column_names = table_column_names.copy()
    #         # source_column_names[0] = f"'{target_scenario_name}'"
    #         # # source_columns_txt = ','.join(source_column_names)
    #         #
    #         # # other_source_column_names =  table_column_names.copy()[1:]  # Drop the scenario_name column
    #         # # other_source_column_names = [f"{db_table.db_table_name}.{n}" for n in other_source_column_names]
    #         # other_source_columns_txt = ','.join(other_source_column_names)
    #         # # source_columns = ','.join(f'"{n}"' for n in source_column_names[1:])
    #         # # source_columns_txt = f"'{target_scenario_name}', {source_columns}"
    #
    #         t: sqlalchemy.table = db_table.table_metadata  # The table at hand
    #         s: sqlalchemy.table = sa_scenario_table  # The scenario table
    #         # print("+++++++++++SQLAlchemy insert-select")
    #         select_columns = [s.c.scenario_name if c.name == 'scenario_name' else c for c in t.columns]  # Replace the t.c.scenario_name with s.c.scenario_name, so we get the new value
    #         # print(f"select columns = {select_columns}")
    #         select_sql = (sqlalchemy.select(select_columns)
    #                       .where(sqlalchemy.and_(t.c.scenario_name == source_scenario_name, s.c.scenario_name == target_scenario_name)))
    #         target_columns = [c for c in t.columns]
    #         sql_insert = t.insert().from_select(target_columns, select_sql)
    #         # print(f"sql_insert = {sql_insert}")
    #
    #         # sql_insert = f"INSERT INTO {db_table.db_table_name} ({target_columns_txt}) SELECT '{target_scenario_name}',{other_source_columns_txt} FROM {db_table.db_table_name} WHERE scenario_name = '{source_scenario_name}'"
    #         if batch_sql:
    #             sql_statements.append(sql_insert)
    #         else:
    #             connection.execute(sql_insert)
    #     if batch_sql:
    #         batch_sql = ";\n".join(sql_statements)
    #         print(batch_sql)
    #         connection.execute(batch_sql)
    #
    #
    # def _find_free_duplicate_scenario_name(self, scenario_name: str, scenarios_df=None) -> Optional[str]:
    #     """Finds next free scenario name based on pattern '{scenario_name}_copy_n'.
    #     Will try at maximum 20 attempts.
    #     """
    #     max_num_attempts = 20
    #     for i in range(1, max_num_attempts + 1):
    #         new_name = f"{scenario_name}({i})"
    #         free = self._check_free_scenario_name(new_name, scenarios_df)
    #         if free:
    #             return new_name
    #     raise ValueError(f"Cannot find free name for duplicate scenario. Tried {max_num_attempts}. Last attempt = {new_name}. Rename scenarios.")
    #     return None
    #
    # def _check_free_scenario_name(self, scenario_name, scenarios_df=None) -> bool:
    #     if scenarios_df is None:
    #         scenarios_df = self.get_scenarios_df()
    #     free = (False if scenario_name in scenarios_df.index else True)
    #     return free
    #
    # ##############################################
    # def rename_scenario_in_db(self, source_scenario_name: str, target_scenario_name: str):
    #     """Rename a scenario. Uses a transaction (when enabled).
    #     TODO: get rename in SQL to work. Currently causing Integrity errors due to not being able to defer constraint checking."""
    #     if self.enable_transactions:
    #         print("Rename scenario within a transaction")
    #         with self.engine.begin() as connection:
    #             # self._rename_scenario_in_db(source_scenario_name, target_scenario_name, connection=connection)
    #             self._rename_scenario_in_db_sql(connection, source_scenario_name, target_scenario_name)
    #     else:
    #         # self._rename_scenario_in_db(source_scenario_name, target_scenario_name)
    #         self._rename_scenario_in_db_sql(self.engine, source_scenario_name, target_scenario_name)
    #
    # # def _rename_scenario_in_db(self, source_scenario_name: str, target_scenario_name: str = None, connection=None):
    # #     """DEPRECATED: use _rename_scenario_in_db_sql: more efficient. Avoids moving data back and forth
    # #     TODO: do this with SQL. Avoids moving data from and too the DB. More efficient.
    # #     :param source_scenario_name:
    # #     :param target_scenario_name:
    # #     :param connection:
    # #     :return:
    # #     """
    # #     if self._check_free_scenario_name(target_scenario_name):
    # #         new_scenario_name = target_scenario_name
    # #     else:
    # #         raise ValueError(f"Target name for rename scenario '{target_scenario_name}' already exists.")
    # #
    # #     # Note that when using a transaction, the order of delete vs insert is irrelevant
    # #     inputs, outputs = self.read_scenario_from_db(source_scenario_name)
    # #     # print(f"KPI columns {outputs['kpis'].columns}")
    # #     outputs['kpis'].columns = outputs['kpis'].columns.str.upper()  # HACK!!! TODO: fix in read_scenario
    # #     self._replace_scenario_in_db_transaction(scenario_name=new_scenario_name, inputs=inputs, outputs=outputs,
    # #                                              bulk=True, connection=connection)
    # #     self._delete_scenario_from_db(scenario_name=source_scenario_name, connection=connection)
    #
    # def _rename_scenario_in_db_sql(self, connection, source_scenario_name: str, target_scenario_name: str = None):
    #     """Rename scenario.
    #     Uses 2 steps:
    #     1. Duplicate scenario
    #     2. Delete source scenario.
    #
    #     Problem is that we use scenario_name as a primary key. You should not change the value of primary keys in a DB.
    #     Instead, first copy the data using a new scenario_name, i.e. duplicate a scenario. Next, delete the original scenario.
    #
    #     Long-term solution: use a scenario_seq sequence key as the PK. With scenario_name as a ordinary column in the scenario table.
    #
    #     Old:
    #     Do a rename using an SQL UPDATE
    #     `UPDATE table SET scenario_name = 'target_scenario_name' WHERE scenario_name = 'source_scenario_name'`
    #     Bottleneck: causes Integrity errors while updating. Need to defer constraint checking!
    #
    #     Work-around 1:
    #     Group all sql UPDATE statements in one sql statement.
    #     -> No exception, but doesn't seem to do anything!
    #     This is anyway NOT a good way
    #     See https://stackoverflow.com/questions/2499246/how-to-update-primary-key
    #     Apparently, we still need to first insert new rows (cascading!), then delete the old rows
    #
    #     Use of 'insert into select': https://stackoverflow.com/questions/9879830/select-modify-and-insert-into-the-same-table
    #     """
    #     # sql_statements = []
    #     # for scenario_table_name, db_table in self.db_tables.items():
    #     #     sql_update = f"UPDATE {db_table.db_table_name} SET SCENARIO_NAME = '{target_scenario_name}' WHERE SCENARIO_NAME = '{source_scenario_name}';"
    #     #     sql_statements.append(sql_update)
    #     # sql = "\n".join(sql_statements)
    #     # print(sql)
    #     # if connection is None:
    #     #     self.engine.execute(sql)
    #     # else:
    #     #     connection.execute(sql)
    #
    #     # 1. Duplicate scenario
    #     self._duplicate_scenario_in_db_sql(connection, source_scenario_name, target_scenario_name)
    #     # 2. Delete scenario
    #     self._delete_scenario_from_db(source_scenario_name, connection=connection)


    # def _delete_scenario_from_db(self, scenario_name: str, connection):
    #     """Deletes all rows associated with a given scenario.
    #     Note that it only deletes rows from tables defined in the self.db_tables, i.e. will NOT delete rows in 'auto-inserted' tables!
    #     Must do a 'cascading' delete to ensure not violating FK constraints. In reverse order of how they are inserted.
    #     Also deletes entry in scenario table
    #     Uses SQLAlchemy syntax to generate SQL
    #     TODO: batch all sql statements in single execute. Faster? And will that do the defer integrity checks?
    #     """
    #     batch_sql=False
    #     insp = sqlalchemy.inspect(connection)
    #     tables_in_db = insp.get_table_names(schema=self.schema)
    #     sql_statements = []
    #     for scenario_table_name, db_table in reversed(self.db_tables.items()):  # Note this INCLUDES the SCENARIO table!
    #         if db_table.db_table_name in tables_in_db:
    #             # sql = f"DELETE FROM {db_table.db_table_name} WHERE scenario_name = '{scenario_name}'"  # Old
    #             t = db_table.table_metadata  # A Table()
    #             sql = t.delete().where(t.c.scenario_name == scenario_name)
    #             if batch_sql:
    #                 sql_statements.append(sql)
    #             else:
    #                 connection.execute(sql)
    #
    #     # Because the scenario table has already been included in above loop, no need to do separately
    #     # Delete scenario entry in scenario table:
    #     # sql = f"DELETE FROM SCENARIO WHERE scenario_name = '{scenario_name}'"
    #     # sql_statements.append(sql)
    #     if batch_sql:
    #         batch_sql = ";\n".join(sql_statements)
    #         print(batch_sql)
    #         connection.execute(batch_sql)

        #############################################################################
        # for scenario_table_name, db_table in reversed(self.db_tables.items()):
        #     # if insp.has_table(db_table.db_table_name, schema=self.schema):  # .has_table() only supported in SQLAlchemy 1.4+
        #     if db_table.db_table_name in tables_in_db:
        #         sql = f"DELETE FROM {db_table.db_table_name} WHERE scenario_name = '{scenario_name}'"
        #         if connection is None:
        #             self.engine.execute(sql)
        #         else:
        #             connection.execute(sql)
        #
        # # Delete scenario entry in scenario table:
        # sql = f"DELETE FROM SCENARIO WHERE scenario_name = '{scenario_name}'"
        # if connection is None:
        #     self.engine.execute(sql)
        # else:
        #     connection.execute(sql)
        #############################################################################
        # if connection is None:
        #     insp = sqlalchemy.inspect(self.engine)
        #     print(f"inspector no transaction = {type(insp)}")
        #     tables_in_db = insp.get_table_names(schema=self.schema)
        #     print(tables_in_db)
        #     for scenario_table_name, db_table in reversed(self.db_tables.items()):
        #         if insp.has_table(db_table.db_table_name, schema=self.schema):
        #         if insp.has_table(db_table.db_table_name, schema=self.schema):
        #             sql = f"DELETE FROM {db_table.db_table_name} WHERE scenario_name = '{scenario_name}'"
        #             self.engine.execute(sql)
        #
        #     # Delete scenario entry in scenario table:
        #     sql = f"DELETE FROM SCENARIO WHERE scenario_name = '{scenario_name}'"
        #     self.engine.execute(sql)
        # else:
        #     insp = sqlalchemy.inspect(connection)
        #     print(f"inspector with transaction= {type(insp)}")
        #     print(insp.get_table_names(schema=self.schema))
        #     for scenario_table_name, db_table in reversed(self.db_tables.items()):
        #         if insp.has_table(db_table.db_table_name, schema=self.schema):
        #             sql = f"DELETE FROM {db_table.db_table_name} WHERE scenario_name = '{scenario_name}'"
        #             connection.execute(sql)
        #
        #     # Delete scenario entry in scenario table:
        #     sql = f"DELETE FROM SCENARIO WHERE scenario_name = '{scenario_name}'"
        #     connection.execute(sql)
    #############################################
    # Migrated VT 2022-01-09:
    # def read_scenario_from_db(self, scenario_name: str, multi_threaded: bool = False) -> (Inputs, Outputs):
    #     """Single scenario load.
    #     Main API to read a complete scenario.
    #     Reads all tables for a single scenario.
    #     Returns all tables in one dict
    #
    #     Fixed: omit reading scenario table as an input.
    #     """
    #     print(f"read_scenario_from_db.multi_threaded = {multi_threaded}")
    #     if multi_threaded:
    #         inputs, outputs = self._read_scenario_from_db_multi_threaded(scenario_name)
    #     else:
    #         if self.enable_transactions:
    #             with self.engine.begin() as connection:
    #                 inputs, outputs = self._read_scenario_from_db(scenario_name, connection)
    #         else:
    #             inputs, outputs = self._read_scenario_from_db(scenario_name, self.engine)
    #     return inputs, outputs

    # Migrated VT 2022-01-09:
    # def _read_scenario_from_db_multi_threaded(self, scenario_name) -> (Inputs, Outputs):
    #     """Reads all tables from a scenario using multi-threading"""
    #     class ReadTableFunction(object):
    #         def __init__(self, dbm):
    #             self.dbm = dbm
    #         def __call__(self, scenario_table_name, db_table):
    #             return self._read_scenario_db_table_from_db_thread(scenario_table_name, db_table)
    #         def _read_scenario_db_table_from_db_thread(self, scenario_table_name, db_table):
    #             with self.dbm.engine.begin() as connection:
    #                 df = self.dbm._read_scenario_db_table_from_db(scenario_name, db_table, connection)
    #                 dict = {scenario_table_name: df}
    #             return dict
    #
    #     thread_number = 8
    #     pool = ThreadPool(thread_number)
    #     thread_worker = ReadTableFunction(self)
    #     print("ThreadPool created")
    #     # input_tables = [(scenario_table_name, db_table) for scenario_table_name, db_table in self.input_db_tables.items()]
    #     # input_results = pool.starmap(thread_worker, input_tables)  # Returns a list of Dict: [{scenario_table_name: df}]
    #     # inputs = {k:v for element in input_results for k,v in element.items()}  # Convert list of Dict to one Dict.
    #     # # print(inputs)
    #     # output_tables = [(scenario_table_name, db_table) for scenario_table_name, db_table in self.output_db_tables.items()]
    #     # output_results = pool.starmap(thread_worker, output_tables)
    #     # outputs = {k:v for element in output_results for k,v in element.items()}
    #
    #     all_tables = [(scenario_table_name, db_table) for scenario_table_name, db_table in self.db_tables.items() if scenario_table_name != 'Scenario']
    #     # print(all_tables)
    #     all_results = pool.starmap(thread_worker, all_tables)
    #     inputs = {k:v for element in all_results for k,v in element.items() if k in self.input_db_tables.keys()}
    #     outputs = {k:v for element in all_results for k,v in element.items() if k in self.output_db_tables.keys()}
    #
    #     print("All tables loaded")
    #
    #     return inputs, outputs


    # Migrated VT 2022-01-09:
    # def _read_scenario_from_db(self, scenario_name: str, connection) -> (Inputs, Outputs):
    #     """Single scenario load.
    #     Main API to read a complete scenario.
    #     Reads all tables for a single scenario.
    #     Returns all tables in one dict
    #     """
    #     inputs = {}
    #     for scenario_table_name, db_table in self.input_db_tables.items():
    #         # print(f"scenario_table_name = {scenario_table_name}")
    #         if scenario_table_name != 'Scenario':  # Skip the Scenario table as an input
    #             inputs[scenario_table_name] = self._read_scenario_db_table_from_db(scenario_name, db_table, connection=connection)
    #
    #     outputs = {}
    #     for scenario_table_name, db_table in self.output_db_tables.items():
    #         outputs[scenario_table_name] = self._read_scenario_db_table_from_db(scenario_name, db_table, connection=connection)
    #         # if scenario_table_name == 'kpis':
    #         #     # print(f"kpis table columns = {outputs[scenario_table_name].columns}")
    #         #     outputs[scenario_table_name] = outputs[scenario_table_name].rename(columns={'name': 'NAME'})  #HACK!!!!!
    #     return inputs, outputs

    # Migrated VT 2022-01-09:
    # def _read_scenario_db_table_from_db(self, scenario_name: str, db_table: ScenarioDbTable, connection) -> pd.DataFrame:
    #     """Read one table from the DB.
    #     Removes the `scenario_name` column.
    #
    #     Modification: based on SQLAlchemy syntax. If doing the plain text SQL, then some column names not properly extracted
    #     """
    #     db_table_name = db_table.db_table_name
    #     # sql = f"SELECT * FROM {db_table_name} WHERE scenario_name = '{scenario_name}'"  # Old
    #     # db_table.table_metadata is a Table()
    #     t = db_table.table_metadata
    #     sql = t.select().where(t.c.scenario_name == scenario_name)  # This is NOT a simple string!
    #     # print(f"_read_scenario_db_table_from_db SQL = {sql}")
    #     # df = pd.read_sql(sql, con=self.engine)
    #     df = pd.read_sql(sql, con=connection)
    #     if db_table_name != 'scenario':
    #         df = df.drop(columns=['scenario_name'])
    #
    #     return df



# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional, Dict

import pandas as pd
import dash
from dash.exceptions import PreventUpdate

from dse_do_dashboard.main_pages.main_page import MainPage
from dash.dependencies import Input, Output, State
from dash import dcc, html
import dash_bootstrap_components as dbc
import pprint
from dash import dash_table

from dse_do_dashboard.main_pages.prepare_data_page import PrepareDataPage
from dse_do_dashboard.utils.dash_common_utils import get_data_table_card_children, get_pivot_table_card_children, \
    diff_dashtable_mi, ScenarioTableSchema, table_type
from dse_do_dashboard.utils.scenariodbmanager_update import DbCellUpdate


class PrepareDataPageEdit(PrepareDataPage):
    """Includes feature to edit input tables.
    Use instead of `PrepareDataPage`.
    Do not combine in same app: will cause duplicate callbacks."""
    def __init__(self, dash_app):
        self.data_table_id = 'input_data_table'
        super().__init__(dash_app,
                         page_name='Prepare Data',
                         page_id='prepare-data',
                         url='prepare-data',
                         )

    def get_layout(self, scenario_name: str = None, reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        input_tables = self.dash_app.get_input_table_names()
        layout = html.Div([
            dcc.Store(id="current_table_name"),

            self.get_input_table_selection_card(),

            dbc.Card([
                # dbc.CardHeader('Input Table'),
                dbc.CardBody(id='input_data_table_card', style={'width': '79vw'},
                             children=get_data_table_card_children(df=pd.DataFrame(), table_name='None', editable=True, data_table_id=self.data_table_id)  # We need to initialize a DataTable, otherwise issues with registering callbacks
                             ),

                html.Button("No table updates", id="commit_changes_button", disabled=True),
                html.Div(id="input_data_table_div"),
                dcc.Store(id="my_data_table_diff_store"), # Stores all changes
                html.Div(id="my_data_table_data_diff"),  # Show the difference
                html.Div(id="my_data_table_output")
            ], style = {'width': '80vw'}),

            self.get_pivot_table_card(),
        ])
        return layout

    def update_data_and_pivot_input_table_callback(self, scenario_name, table_name, diff_store_data=None):
        """Body for the Dash callback.
        Usage::
            @app.callback([Output('input_data_table_card', 'children'),
            Output('input_pivot_table_card', 'children')],
            [Input('top_menu_scenarios_drpdwn', 'value'),
            Input('input_table_drpdwn', 'value')])
            def update_data_and_pivot_input_table(scenario_name, table_name):
                data_table_children, pivot_table_children = DA.update_data_and_pivot_input_table_callback(scenario_name, table_name)
                return [data_table_children, pivot_table_children]

        TODO: share parts with parent
        """
        # print(f"update_data_and_pivot_input_table for {table_name} in {scenario_name}")
        input_table_names = [table_name]
        pm = self.dash_app.get_plotly_manager(scenario_name, input_table_names, [])
        dm = pm.dm
        df = self.dash_app.get_table_by_name(dm=dm, table_name=table_name, index=False, expand=False)
        table_schema = self.dash_app.get_table_schema(table_name)
        pivot_table_config = self.dash_app.get_pivot_table_config(table_name)
        data_table_children = self.get_data_table_card_children(df, table_name, table_schema, editable=True, data_table_id=self.data_table_id, diff_store_data=diff_store_data)
        pivot_table_children = get_pivot_table_card_children(df, scenario_name, table_name, pivot_table_config)
        return data_table_children, pivot_table_children

    def get_data_table_card_children(self, df, table_name:str, table_schema: Optional[ScenarioTableSchema] = None,
                                     editable: bool = False, data_table_id:str=None, diff_store_data=None):
        return [
            dbc.CardHeader(
                table_name
                # title=table_name,
                # fullscreen=True
            ),
            self.get_data_table(df, table_schema, editable, data_table_id, diff_store_data)
        ]

    def get_data_table(self, df, table_schema: Optional[ScenarioTableSchema] = None, editable: bool = False, data_table_id=None, diff_store_data=None) -> dash_table.DataTable:
        """
        Generates a DataTable for a DataFrame. For use in 'Prepare Data' and 'Explore Solution' pages.
        :param df:
        :param table_schema:
        :return:
        """
        if data_table_id is None:
            data_table_id = 'my_data_table'
        index_columns = []
        if table_schema is not None:
            # print(f"prepare_data_page_edit - get_data_table table_schema.index_columns={table_schema.index_columns}")
            if len(table_schema.index_columns) > 0:
                df = df.set_index(table_schema.index_columns).reset_index()  # ensures all index columns are first
            index_columns = table_schema.index_columns
        return dash_table.DataTable(
            id=data_table_id,
            data=df.to_dict('records'),
            columns=[
                {'name': i, 'id': i, 'type': table_type(df[i])}
                for i in df.columns
            ],
            fixed_rows={'headers': True},
            editable=editable,
            # fixed_columns={'headers': False, 'data': 0}, # Does NOT create a horizontal scroll bar
            filter_action="native",
            sort_action="native",
            sort_mode="multi",
            style_cell={
                'textOverflow': 'ellipsis',  # See https://dash.plotly.com/datatable/width to control column-name width
                'maxWidth': 0,               # Needs to be here for the 'ellipsis' option to work
                'overflow' : 'hidden',
                'font_family': 'sans-serif',
                'font_size': '12px',
                'textAlign': 'left'},
            style_table={
                'maxHeight': '400px',
                'overflowY': 'scroll'
            },
            style_header={
                'if': {
                    'column_id': index_columns
                },
                # 'backgroundColor': 'rgb(230, 230, 230)',
                'fontWeight': 'bold'
            },
            style_data_conditional=([
                {
                    'if': {
                        'column_id': index_columns
                    },
                    'fontWeight': 'bold',
                    # 'backgroundColor': '#0074D9',
                    # 'color': 'white'
                },
                # {
                #     'if': {
                #         'row_index': 2,
                #         'column_id': 'size'
                #     },
                #     'backgroundColor': '#d62728'
                # },
            ] + self.get_table_style_data_conditional_cell_edit(diff_store_data)
            ))

    def get_table_style_data_conditional_cell_edit(self, diff_store_data=None) -> List[Dict]:
        """Return a list of dicts with conditional cell style. For each cell that has been edited.
        Partly works:
        - The coloring works, but
        - The table needs to be refreshed/regenerated to get the color filtering active.
        That refresh is not happening currently. Need to see how it can be triggered
        That refresh will reset the edits in the table
        So one would need to re-apply the changes from the store.
        Seems like a lot of work for this feature.
        """
        # print(f"get_table_style_data_conditional_cell_edit diff_store_data = {diff_store_data}")
        if diff_store_data is None:
            return []
        cell_edits = []
        for l in diff_store_data.values():
            for diff in l:
                cell_edits.append({
                    'if': {
                        'row_index': diff['row_idx'],
                        'column_id': diff['column_name']
                    },
                    'backgroundColor': '#d62728'
                })
        return cell_edits


    def get_db_cell_updates(self, diff_store_data, ) -> List[DbCellUpdate]:
        """Get the changes in the diff store as DbCellUpdate.
        Note: from Python 3.7, a Dict maintains the order in which items are added. So no need to sort by time.
        TODO: can we store the NamedTuple `DbCellUpdate` directly in the store? Saves us from converting from Dict.
        """
        db_cell_updates: List[DbCellUpdate] = []
        for ts, diffs in diff_store_data.items():  # Potentially multiple changes in the value
            for diff in diffs:
                db_cell_update = DbCellUpdate(
                    scenario_name=diff['scenario_name'],
                    table_name=diff['table_name'],
                    row_index=diff['row_index'],  # e.g. [{'column': 'col1', 'value': 1}, {'column': 'col2', 'value': 'pear'}]
                    column_name=diff['column_name'],
                    current_value=diff['current_value'],
                    previous_value=diff['previous_value'],
                    row_idx=diff['row_idx'],
                )
                db_cell_updates.append(db_cell_update)
        return db_cell_updates

    def capture_diffs_callback(self, ts, data, data_previous, diff_store_data, table_name: str, scenario_name: str):
        """Capture diffs and store in my_data_table_diff_store."""
        if ts is None:
            raise PreventUpdate
        diff_store_data = diff_store_data or {}
        index_columns = None
        table_schema: ScenarioTableSchema = self.dash_app.get_table_schema(table_name)
        if table_schema is not None:
            index_columns = table_schema.index_columns
        diff_store_data[ts] = diff_dashtable_mi(data, data_previous, index_columns=index_columns, table_name=table_name, scenario_name=scenario_name)
        return diff_store_data

    def commit_changes_to_db_callback(self, n_clicks, diff_store_data):
        if n_clicks is None:
            raise PreventUpdate
        if diff_store_data is None:
            raise PreventUpdate
        if diff_store_data:
            dt_changes = []
            for ts, v in diff_store_data.items():
                dt_changes.append(f"* {v} {ts}")
            # for v in diff_store_data.values():
            #     dt_changes.append(f"* {v}")
            output = [dcc.Markdown(change) for change in dt_changes]
        else:
            output = "No Changes to DataTable"

        db_cell_updates = self.get_db_cell_updates(diff_store_data)
        self.dash_app.dbm.update_cell_changes_in_db(db_cell_updates)

        return output

    def set_dash_callbacks(self):
        """Define Dash callbacks for this page

        Will be called to register any callbacks
        :return:
        """
        # print("Set dash callbacks for PrepareDataPageEdit")
        # super().set_dash_callbacks()
        app = self.dash_app.app

        @app.callback([Output('input_data_table_card', 'children'),
                       Output('input_pivot_table_card', 'children')],
                      [Input('top_menu_scenarios_drpdwn', 'value'),
                       Input('input_table_drpdwn', 'value')],
                      State("my_data_table_diff_store", "data"))
        def update_data_and_pivot_input_table_edit(scenario_name:str, table_name:str, diff_store_data):
            data_table_children, pivot_table_children = self.update_data_and_pivot_input_table_callback(scenario_name, table_name, diff_store_data)
            return [data_table_children, pivot_table_children]

        @app.callback([Output('commit_changes_button', 'style'),
                       Output('commit_changes_button', 'children')
                       ],
                      Input("my_data_table_diff_store", "data"))
        def change_button_style(data):
            if data is None:
                raise PreventUpdate

            white_button_style = {'background-color': 'white',
                                  'color': 'black',
                                  # 'height': '50px',
                                  # 'width': '100px',
                                  # 'margin-top': '50px',
                                  # 'margin-left': '50px'
                                  }

            red_button_style = {'background-color': 'red',
                                'color': 'white',
                                # 'height': '50px',
                                # 'width': '100px',
                                # 'margin-top': '50px',
                                # 'margin-left': '50px'
                                }

            if len(data) > 0:
                return red_button_style, f"Commit {len(data)} changes to DB"
            else:
                return white_button_style, "No changes to commit"

        # print("Set dash callbacks for PrepareDataPageEdit.capture_and_commit_edits")
        @app.callback(
            [
             Output("my_data_table_diff_store", "data"),
             Output('commit_changes_button', 'disabled')],
            [Input(self.data_table_id, "data_timestamp"),
             Input("commit_changes_button", "n_clicks")],
            [
                State(self.data_table_id, "data"),
                State(self.data_table_id, "data_previous"),
                State("my_data_table_diff_store", "data"),
                State("input_table_drpdwn", "value"),
                State('top_menu_scenarios_drpdwn', 'value'),
            ],
        )
        def capture_and_commit_edits(ts, n_clicks, data, data_previous, diff_store_data, table_name: str, scenario_name: str):
            """Combined callback for either editing cells in the DataTable, or pressing the commit button.
            Editing a cell detects and stores changes in the diff store.
            The commit button performs updates in the DB and clears the diff store.
            """
            ctx = dash.callback_context
            if not ctx.triggered:
                # print(f"Not triggered. diff_store_data = {diff_store_data}")
                # raise PreventUpdate
                # Attempt to enable Commit button if there are changes pending in the diff-store after a 'Refresh' button click
                # Note that the value of the change in the table has been lost due to the refresh, but not the Store
                # So the diff can still be committed, but doesn't properly show in the table
                if diff_store_data is None:
                    raise PreventUpdate
                else:
                    return diff_store_data, False

            triggered_component_id = ctx.triggered[0]['prop_id'].split('.')[0]

            if triggered_component_id == self.data_table_id:
                # print("Triggered by a table edit")
                diff_store_data = self.capture_diffs_callback(ts,data, data_previous, diff_store_data, table_name, scenario_name)
                commit_button_disabled = False
            elif triggered_component_id == 'commit_changes_button':
                # print("Triggered by commit button")
                output = self.commit_changes_to_db_callback(n_clicks, diff_store_data)
                diff_store_data = {}  # Clear diff store
                commit_button_disabled = True
            else:
                print(f"Triggered by unknown component {triggered_component_id}")
                raise PreventUpdate

            return diff_store_data, commit_button_disabled


        # @app.callback(
        #     [Output("my_data_table_diff_store", "data"),
        #      Output('commit_changes_button', 'disabled')],
        #     [Input("my_data_table", "data_timestamp")],
        #     [
        #         State("my_data_table", "data"),
        #         State("my_data_table", "data_previous"),
        #         State("my_data_table_diff_store", "data"),
        #         State("input_table_drpdwn", "value"),
        #         State('top_menu_scenarios_drpdwn', 'value'),
        #     ],
        # )
        # def capture_diffs(ts, data, data_previous, diff_store_data, table_name: str, scenario_name: str):
        #     """Capture diffs and store in my_data_table_diff_store."""
        #     if ts is None:
        #         raise PreventUpdate
        #     diff_store_data = diff_store_data or {}
        #     index_columns = None
        #     table_schema: ScenarioTableSchema = self.dash_app.get_table_schema(table_name)
        #     if table_schema is not None:
        #         index_columns = table_schema.index_columns
        #     diff_store_data[ts] = diff_dashtable_mi(data, data_previous, index_columns=index_columns, table_name=table_name, scenario_name=scenario_name)
        #     return diff_store_data, False

        # @app.callback(
        #     Output("my_data_table_data_diff", "children"),
        #     # Output("my_data_table_diff_store", "data"),
        #     [Input("commit_changes_button", "n_clicks")],
        #     [State("my_data_table_diff_store", "data")],
        # )
        # def update_output(n_clicks, diff_store_data):
        #     if n_clicks is None:
        #         raise PreventUpdate
        #     if diff_store_data is None:
        #         raise PreventUpdate
        #     if diff_store_data:
        #         dt_changes = []
        #         for ts, v in diff_store_data.items():
        #             dt_changes.append(f"* {v} {ts}")
        #         # for v in diff_store_data.values():
        #         #     dt_changes.append(f"* {v}")
        #         output = [dcc.Markdown(change) for change in dt_changes]
        #     else:
        #         output = "No Changes to DataTable"
        #
        #     db_cell_updates = self.get_db_cell_updates(diff_store_data)
        #     self.dash_app.dbm.update_cell_changes_in_db(db_cell_updates)
        #
        #     diff_store_data = {}
        #
        #     return output #, diff_store_data

###########################################################################
# Test to use the dcc.Location to stor the table name as a html variable
# Disadvantage/problem is that a change to the url causes a page-reload
###########################################################################
        # @app.callback(Output("my_data_table_data_diff", "children"),
        #               Input('url', 'search'))
        # def test_url_vars(search):
        #     print(f"test_url_vars = {search}")
        #     return "Test"
        #
        # @app.callback(Output('url', 'search'),
        #               Input('input_table_drpdwn', 'value'),
        #               prevent_initial_call=True)
        # def test_url_vars_output(table_name):
        #     return f"?table={table_name}"

##################################################################################
# Test using a dcc.Store to store the table_name, so switching scenarios keeps the same table
# Not straightforward: the callback update_data_and_pivot_input_table would have to do a context test to see which input triggered the callback
# Also, not sure the store gets retained when switching scenarios
# (surprisingly, that DOES seem to be the case for the diff-store!)
##################################################################################
        # @app.callback(Output('current_table_name', 'data'),
        #               Input('input_table_drpdwn', 'value'))
        # def store_table_name(table_name):
        #     return table_name
        #
        # @app.callback(Output("my_data_table_data_diff", "children"),
        #               Input('current_table_name', 'data'))
        # def test_table_name(data):
        #     return data



from nbformat import read, NO_CONVERT
from io import StringIO
from contextlib import redirect_stdout
import sys
import traceback
from typing import Dict


class InterpreterError(Exception):
    pass


def my_exec(cmd, globals=None, locals=None, description='source string'):
    """Does an `exec()` with exception handling to get the line-number where the exception has occurred in the extracted code (cmd)."""
    try:
        exec(cmd, globals, locals)
    except SyntaxError as err:
        error_class = err.__class__.__name__
        detail = err.args[0]
        line_number = err.lineno
    except Exception as err:
        error_class = err.__class__.__name__
        detail = err.args[0]
        cl, exc, tb = sys.exc_info()
        line_number = traceback.extract_tb(tb)[-1][1]
    else:
        return
    raise InterpreterError("%s at line %d of %s: %s" % (error_class, line_number, description, detail))


class NotebookRunner():
    """Runs the code from a notebook.
    Extracts code from notebook.
    Runs the code. Returns the log (i.e. stdoutput).

    Usage::

        runner = NotebookRunner(notebook_path)
        log = runner.run(globals(), locals())
        print(log)

    """

    def __init__(self, notebook_path: str, exec_globals: Dict = {}, exec_locals: Dict  = {}):
        """Instantiate a NotebookRunner.
        :param (str): file_path of notebook
        :param exec_globals (Dict): set of global variables we want to expose to the exec code. E.g. `{'inputs':inputs}`
        Typically, for a DO model, it should include the inputs dictionary.
        :param exec_locals (Dict): typically nothing/empty dict. Any local variable that gets assigned will become available in self.exec_locals
        """
        self.notebook_path = notebook_path
        self.code = ""
        self.log = ""
        self.exec_globals = exec_globals
        self.exec_locals = exec_locals

    def run(self) -> str:
        """Extracts and runs the notebook.
        Returns the log.
        """
        self.code = self._get_nb_code()
        self.log = self._run_code(self.code)
        return self.log

    def _get_nb_code(self):
        """
        WEIRD: for some crazy unknown reason, typing (lowercase) 'CELLS' causes CPD to fail writing a .py file!
        Anywhere in the file, including comments!
        So below I try to avoid this term, which is why I have to do the:
        * 'cellz'
        * 'lower()' operation on 'CELLS'.
        """
        with open(self.notebook_path) as fp:
            notebook = read(fp, NO_CONVERT)
            cellz = notebook['CELLS'.lower()]
            code_cellz = [c for c in cellz if c['cell_type'] == 'code']
            code = ""
            for cell in code_cellz:
                cell_source = cell['source']
                if self.include_code_cell(cell_source):
                    code += cell_source + "\n"
        return code

    def include_code_cell(self, cell_source):
        """Default implementation always returns True.
        To be overridden for custom cell filtering."""
        return True

    def _run_code(self, code):
        ccode = compile(code, "somefile.py", 'exec')  # Compile the code
        f = StringIO()  # For grabbing the log
        with redirect_stdout(f):
            # exec(ccode)
            try:
                my_exec(ccode, self.exec_globals, self.exec_locals)  # Execute the compiled Python code
            except Exception as err:
                print("Exception")
                print(err)  # TODO: more info
        s = f.getvalue()  # get the log
        return s


class DoNotebookRunner(NotebookRunner):
    """Runs the code extracted from a 'DO solver' notebook designed to work with a DO Experiment.
    Ignores any code cell starting with `#dd-ignore`.
    A DoNotebookRunner assumes an input dictionary is passed-in as the only 'globals'.
    And all outputs are added to the dictionary 'outputs'.
    An empty `outputs` dictionary is part of the globals, so the code can either add output DataFrames to the existing outputs,
    or re-create a new outputs dictionary.

    For now, add globals() and locals() in constructor, otherwise there are issues with importing certain packages.
    The `inputs` and an empty `outputs` will be added in each run.

    Usage::

        runner = DoNotebookRunner(notebook_path, globals(), locals())
        outputs = runner.run(inputs)

    """

    #     def __init__(self, notebook_path):
    #         super().__init__(notebook_path)

    def run(self, inputs: Dict = {}) -> Dict:
        """Runs the notebook.
        Passes-in the inputs as part of the exec_globals.
        And adds an empty `outputs` to the globals, so that the notebook can add output DataFrames, e.g. `outputs["a"] = xxx`.
        Notebook is also allowed to re-create the outputs Dict, e.g. `outputs = {"a": xxx}`
        """
        self.exec_globals['inputs'] = inputs
        self.exec_globals['outputs'] = {}
        super().run()
        return self._get_outputs()

    def include_code_cell(self, cell_source):
        return not cell_source.startswith("#dd-ignore")

    def _get_outputs(self):
        """Get the outputs dictionary.
        If the code has re-assigned the outputs (i.e. `outputs = xxxxx`), then the outputs will be in the exec_locals.
        An empty outputs has also been passed in the globals, so the code could also just have added key-value pairs, e.g. `outputs["a"] = xxxx`.
        Then the outputs will be available in the exec_globals.
        """
        if 'outputs' in self.exec_locals:
            outputs = self.exec_locals['outputs']
        elif 'outputs' in self.exec_globals:
            outputs = self.exec_globals['outputs']
        else:
            outputs = {}
        return outputs

    def print_code(self):
        def insert_line_numbers(txt):
            return "\n".join([f"{n+1:03d} {line}" for n, line in enumerate(txt.split("\n"))])
        print(insert_line_numbers(self.code))
        # line_counter = 1
        # for line in self.code:
        #     print(f"{line_counter} - {line}")
        #     line_counter += 1




# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import List

from dse_do_dashboard.main_pages.main_page import MainPage
from dash import dcc, html, Output, Input, State
import dash_daq as daq
import dash_bootstrap_components as dbc


class HomePage(MainPage):
    def __init__(self, dash_app):
        super().__init__(dash_app,
                         page_name='Home',
                         page_id='home',
                         url='',
                         )

    def get_layout(self, scenario_name: str = None, reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        scenarios_df = self.dash_app.read_scenarios_table_from_db_cached().reset_index()  # SCDB2.get_scenarios_df().reset_index()
        layout = html.Div([

            dbc.Card([
                dbc.CardHeader(html.Div("Reference Scenario", style={'width': '28vw'})),
                dbc.CardBody([
                    # dbc.CardHeader(html.Div("Reference Scenario", style={'width': '28vw'})),
                    dcc.Dropdown(
                        id='reference_scenario',
                        options=[
                            {'label': i, 'value': i}
                            for i in scenarios_df.scenario_name
                        ],  style = {'width': '28vw'})
                ])
            ], style = {'width': '30vw'}),

            dbc.Card([
                dbc.CardBody(
                    # id='scenario_table_card',
                    [dbc.CardHeader(html.Div("Scenarios", style={'width': '80vw'})),
                     html.Div(id="scenario_table_div", style={'width': '78vw'})]
                ),
            ], style={'width': '80vw'}),

            dbc.Card([
                dbc.CardBody(
                    [
                        dbc.Button(
                            "Show 'Stop Server' Button",
                            id="collapse_stop_server_button",
                            className="mb-3",
                            color="primary",
                            n_clicks=0,
                        ),
                        dbc.Collapse(
                            dbc.Card(dbc.CardBody([
                                daq.StopButton(
                                    id='stop_server_button',
                                    label=f'Stop the Dash server. '
                                          f'Will release the port number {self.dash_app.port}.',
                                    n_clicks=0
                                ),
                                html.Div(id='stop_server_button_output')
                            ])),
                            id="collapse_stop_server_button_state",
                            is_open=False,
                        ),
                    ]
                ),
            ], style={'width': '80vw'}),

        ])
        return layout

    def set_dash_callbacks(self):
        app = self.dash_app.app

        @app.callback(
            Output("collapse_stop_server_button_state", "is_open"),
            [Input("collapse_stop_server_button", "n_clicks")],
            [State("collapse_stop_server_button_state", "is_open")],
        )
        def toggle_collapse(n, is_open):
            if n:
                return not is_open
            return is_open

        @app.callback(
            Output('stop_server_button_output', 'children'),
            Input('stop_server_button', 'n_clicks'),
            prevent_initial_call=True
        )
        def update_output(n_clicks):
            self.dash_app.shutdown()
            return 'The stop button has been clicked {} times.'.format(n_clicks)


# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Dict, List

from dse_do_dashboard.main_pages.main_page import MainPage
from dash import dcc, html, Output, Input, State
import dash_bootstrap_components as dbc

from dse_do_dashboard.visualization_pages.visualization_page import VisualizationPage


class VisualizationTabsPage(MainPage):
    """Visualization pages in tabbed layout

    TODO (minor): handle case when there are no visualization pages (i.e. len(self.dash_app.visualization_pages) == 0)
    """
    def __init__(self, dash_app):
        super().__init__(dash_app,
                         page_name='Visualization',
                         page_id='visualization',
                         url='visualization',
                         )

    def get_layout(self, scenario_name: str = None, reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        tab_style = {"fontWeight": "bold"}
        visualization_pages = self.dash_app.visualization_pages
        tab_children = [
            dcc.Tab(
                label=vp.page_name,
                value=vp.page_id,
                style=tab_style,
                selected_style=tab_style,
            ) for vp in visualization_pages
        ]
        tabs = dcc.Tabs(
            id="tabs",
            value=visualization_pages[0].page_id,
            children=tab_children,
        )
        layout = html.Div([
            dbc.Card(children=tabs),
            html.Div(id='page_content'),
        ]
        )
        return layout

    def get_visualization_tab_layout_callback(self, page_id, scenario_name: str, reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        """
        Callback to render the visualization tabs.
        Call from DoDashApp, which is in turn is called from index.py::

            @app.callback(Output("page_content", "children"),
              [Input("tabs", "value"),
               Input('top_menu_scenarios_drpdwn', 'value')])
            def get_visualization_tab_layout_callback(page_id, scenario_name):
                return DA.get_visualization_tab_layout_callback(page_id, scenario_name)

        """
        print(f"get_visualization_tab_layout_callback for {page_id}")
        visualization_pages_dict: Dict[str, VisualizationPage] = {vp.page_id: vp for vp in self.dash_app.visualization_pages}
        if page_id in visualization_pages_dict:
            vp = visualization_pages_dict[page_id]
            try:
                print(f"Getting layout for {page_id}")
                # tab_layout = getattr(sys.modules[f"visualization_pages.{vp.module_name}"], 'layout')
                tab_layout = vp.get_layout(scenario_name, reference_scenario_name, multi_scenario_names)
                return tab_layout

            except KeyError:
                print(f"ERROR in visualization render_tabs2: unknown module_name = {vp.module_name}")
                # return main_pages.not_found.layout
                return self.dash_app.get_not_found_page().get_layout(f"ERROR in visualization render_tabs2: unknown module_name = {vp.module_name}")

        else:
            print(f"ERROR in visualization render_tabs2: unknown page_id = {page_id}")
            # return main_pages.not_found.layout
            return self.dash_app.get_not_found_page().get_layout(f"ERROR in visualization render_tabs2: unknown page_id = {page_id}")

    def set_dash_callbacks(self):
        """Define Dash callbacks for this page

        Will be called to register any callbacks
        :return:
        """
        app = self.dash_app.app

        @app.callback(Output("page_content", "children"),
                      [Input("tabs", "value"),
                       Input('top_menu_scenarios_drpdwn', 'value')],
                      [State('reference_scenario_name_store', 'data'),
                       State('multi_scenario_names_store', 'data')])
        def get_visualization_tab_layout_callback(page_id, scenario_name, reference_scenario_name, multi_scenario_names):
            """To update the tabbed-visualization page."""
            print("xxxxxxxxxxxxxxxxxxxxxx")
            return self.get_visualization_tab_layout_callback(page_id, scenario_name,
                                                              reference_scenario_name=reference_scenario_name,
                                                              multi_scenario_names=multi_scenario_names)



# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import List, NamedTuple

from dash import dcc
import dash_bootstrap_components as dbc
import plotly.graph_objs as go

from dse_do_dashboard.do_dash_app import DoDashApp
from dse_do_dashboard.visualization_pages.visualization_page import VisualizationPage
from dse_do_utils.plotlymanager import PlotlyManager

# from fruit_distribution.fruit.fruitplotlymanager import FruitPlotlyManager


class KpiPageTemplate(VisualizationPage):
    """Abstract class. Creates a layout of KPI gauges with multiple rows of KPI gauges.
    Override the method `get_plotly_figures`.
    """
    def __init__(self, dash_app: DoDashApp, page_name: str = 'Default', page_id: str = 'default', url: str = 'default',
                 input_table_names: List[str] = [], output_table_names: List[str] = []):
        super().__init__(dash_app=dash_app,
                         page_name=page_name,
                         page_id=page_id,
                         url=url,
                         input_table_names=input_table_names,
                         output_table_names=output_table_names,
                         )

    def get_plotly_figures(self, pm: PlotlyManager) -> List[List[go.Figure]]:
        """Create gauges for KPI page as rows of columns.
        That is, each entry in the gauges represents a row. Each row contains a set of go.Figure.

        Abstract method. This is just a sample of a 2x2 matrix of KPIs. Override to """
        kpi_df = pm.dm.kpis
        gauges = [
            [
                KpiPageTemplate.make_traffic_light_gauge(
                    title="KPI 1 %",
                    value=round(.88 * 100, 3),
                    orange_threshold=75,
                    red_threshold=90,
                    max_val=100),
                KpiPageTemplate.make_traffic_light_gauge(
                    title="KPI 2",
                    value=55,
                    orange_threshold=75,
                    red_threshold=90,
                    max_val=100),
            ],
            [
                KpiPageTemplate.make_traffic_light_gauge(
                    title="KPI 3 %",
                    value=round(.33 * 100, 3),
                    orange_threshold=25,
                    red_threshold=50,
                    max_val=100),
                KpiPageTemplate.make_traffic_light_gauge(
                    title="KPI 4",
                    value=777,
                    orange_threshold=750,
                    red_threshold=900,
                    max_val=1000),
            ],
        ]
        return gauges

    def get_layout_children(self, pm: PlotlyManager):
        """Creates a layout for KPI gauges with multiple rows and columns.
        Each row can contain any number of columns."""
        figures = self.get_plotly_figures(pm=pm)

        figure_style = {'height': '20vh', 'width': '20vw', 'margin-left': 'auto', 'margin-right': 'auto',
                       'display': 'block'}
        layout_childen = []

        for figure_row in figures:
            row_layout = dbc.Row([
                dbc.Col(
                    dbc.Card([
                        dbc.CardBody(
                            dcc.Graph(
                                style=figure_style,
                                figure=figure
                            )
                        )
                    ])
                )
                for figure in figure_row
            ])
            layout_childen.append(row_layout)

        return layout_childen

    @staticmethod
    def make_traffic_light_gauge(value: float, title: str, orange_threshold: float, red_threshold: float, max_val: float):
        """
        """
        steps = [
            {'range': [0, orange_threshold], 'color': 'green'},
            {'range': [orange_threshold, red_threshold], 'color': 'orange'},
            {'range': [red_threshold, max_val], 'color': 'red'},
        ]

        fig = go.Figure(go.Indicator(
            mode = "gauge+number",
            value = value,
            domain = {'x': [0, 1], 'y': [0, .75]},
            title = {'text': title, 'font': {'color': 'black', 'size': 18}},
            gauge = {'axis': {'range': [None, max_val], 'tickfont': {'color': 'black'}},
                     'threshold' : {'line': {'color': "darkred", 'width': 4}, 'thickness': 0.75, 'value': red_threshold},
                     'steps': steps,
                     'bar': {'color': "darkblue"},},
        ))

        fig.update_layout(font = {'color': 'green' if value < orange_threshold else 'orange' if value > orange_threshold and value < red_threshold else 'red', 'family': "Arial"},
                          margin={'t':10,'b':30},
                          )

        return fig


# Copyright IBM All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import ast
import base64
import io
import os
import pathlib
import tempfile
import zipfile
from typing import List

import flask
import pandas as pd
from dash.exceptions import PreventUpdate
import dash
from dse_do_utils import ScenarioManager

from dse_do_dashboard.main_pages.main_page import MainPage
from dash import dcc, html, Output, Input, State, ALL, MATCH
import dash_daq as daq
import dash_bootstrap_components as dbc


class HomePageEdit(MainPage):
    """
    Includes:
    - Duplicate Scenario
    - Rename Scenario
    - Delete Scenario
    - Download Scenario(s)
    - Upload Scenario(s)
    """
    def __init__(self, dash_app):
        super().__init__(dash_app,
                         page_name='Home',
                         page_id='home',
                         url='',
                         )

    def get_layout(self, scenario_name: str = None, reference_scenario_name: str = None, multi_scenario_names: List[str] = None):
        scenarios_df = self.dash_app.read_scenarios_table_from_db_cached()  #.reset_index()  # SCDB2.get_scenarios_df().reset_index()

        # print(f"get_layout ref_scenario={reference_scenario_name} , ms={multi_scenario_names}")
        selected_ref_scenarios = [] if multi_scenario_names is None else multi_scenario_names

        # Reconcile reference_scenario_name and multi_scenario_names with actual scenarios? (if in the mean time a scenario got deleted)
        selected_ref_scenario = reference_scenario_name if reference_scenario_name in scenarios_df.index else None
        selected_ref_scenarios = [i for i in selected_ref_scenarios if i in scenarios_df.index]

        layout = html.Div([

            dbc.Card([
                dbc.CardHeader(html.Div("Reference Scenario", style={'width': '28vw'})),
                dbc.CardBody([
                    # dbc.CardHeader(html.Div("Reference Scenario", style={'width': '28vw'})),
                    dcc.Dropdown(
                        id='reference_scenario_drpdwn',
                        options=[
                            {'label': i, 'value': i}
                            # for i in scenarios_df.reset_index().scenario_name
                            for i in scenarios_df.index
                        ],
                        value=selected_ref_scenario,
                        style = {'width': '28vw'})
                ])
            ], #style = {'width': '30vw'}
            ),

            dbc.Card([
                dbc.CardHeader(html.Div("Reference Scenarios", style={'width': '28vw'})),
                dbc.CardBody([
                    # dbc.CardHeader(html.Div("Reference Scenario", style={'width': '28vw'})),
                    dcc.Checklist(
                        id='reference_scenarios_checklist',
                        options=[
                            {'label': i, 'value': i}
                            # for i in scenarios_df.reset_index().scenario_name
                            for i in scenarios_df.index
                        ],
                        value=selected_ref_scenarios,
                        labelStyle={'display': 'block'},
                        # style={"overflow":"auto"}
                        # style = {'width': '28vw'}
                    )
                ])
            ], #style = {'width': '30vw'}
            ),

            dbc.Card([
                dbc.CardHeader(html.Div("Scenarios", style={'width': '80vw'})),
                dbc.CardBody(
                    # id='scenario_table_card',
                    [
                     # html.Div(id="scenario_table_div", style={'width': '78vw'},
                     #          ),
                     html.Div(children=self.get_scenario_operations_table(scenarios_df)),
                     ]
                ),
            ], # style={'width': '80vw'}
            ),

            dbc.Accordion(
                [
                    dbc.AccordionItem(
                        [
                            html.P("Download all scenarios in a .zip archive. May take a long time."),
                            html.Hr(),
                            dbc.Button(
                                "Download all scenarios",
                                id="download_scenarios_button",
                                className="mb-3",
                                color="primary",
                                n_clicks=0,
                            ),
                            dcc.Download(id='download_scenarios_download'),
                        ],
                        title="Download All Scenarios",
                    ),

                    dbc.AccordionItem(
                        [
                            html.P("Select or drop one or more .xlsx or one .zip (with multiple .xlsx)"),
                            html.Hr(),
                            dcc.Upload(
                                id='upload_scenario', # 'upload-data',
                                children=html.Div([
                                    'Drag and Drop or ',
                                    html.A('Select Files')
                                ]),
                                style={
                                    'width': '100%',
                                    'height': '60px',
                                    'lineHeight': '60px',
                                    'borderWidth': '1px',
                                    'borderStyle': 'dashed',
                                    'borderRadius': '5px',
                                    'textAlign': 'center',
                                    'margin': '10px'
                                },
                                # Allow multiple files to be uploaded
                                multiple=True
                            ),
                            html.Div(id='output_data_upload'),
                        ],
                        title="Upload Scenarios",
                    ),

                    dbc.AccordionItem(
                        [
                            daq.StopButton(
                                id='stop_server_button',
                                label=f'Stop the Dash server. '
                                      f'Will release the port number {self.dash_app.port}.',
                                n_clicks=0
                            ),
                            html.Div(id='stop_server_button_output')
                        ],
                        title="Stop Server",
                    ),
                ],
                start_collapsed=True,
            ),

            # dbc.Card([
            #     dbc.CardBody(
            #         [
            #             dbc.Button(
            #                 "Show 'Stop Server' Button",
            #                 id="collapse_stop_server_button",
            #                 className="mb-3",
            #                 color="primary",
            #                 n_clicks=0,
            #             ),
            #             dbc.Collapse(
            #                 dbc.Card(dbc.CardBody([
            #                     # daq.StopButton(
            #                     #     id='stop_server_button',
            #                     #     label=f'Stop the Dash server. '
            #                     #           f'Will release the port number {self.dash_app.port}.',
            #                     #     n_clicks=0
            #                     # ),
            #                     html.Div(id='stop_server_button_output')
            #                 ])),
            #                 id="collapse_stop_server_button_state",
            #                 is_open=False,
            #             ),
            #         ]
            #     ),
            # ], style={'width': '80vw'}),

        ])
        return layout

    def get_scenario_operations_table(self, scenarios_df, ):
        """Create a layout which allows a user to select an operation on a scenario: duplicate, rename, delete"""
        layout = []
        # for scenario_name in scenarios_df.reset_index().scenario_name:
        for scenario_name in scenarios_df.index:
            layout.append(
                dbc.Card(dbc.Row([
                    dbc.Col(self.get_scenario_edit_dropdown(scenario_name, scenarios_df), width=1),
                    dbc.Col(scenario_name),])),
            )
        # layout.extend([
        #     dbc.Button(
        #         "Download all scenarios",
        #         id="download_scenarios_button",
        #         className="mb-3",
        #         color="primary",
        #         n_clicks=0,
        #     ),
        #     dcc.Download(id='download_scenarios_download'),
        # ]
        # )
            # break
        return layout

    def get_scenario_edit_dropdown(self, scenario_name, scenarios_df):
        dropdown = html.Div(
            [
                dbc.DropdownMenu(
                    [
                        dbc.DropdownMenuItem(
                            "Duplicate",
                            id = {'type':'duplicate_scenario_mi', 'index': scenario_name},
                            n_clicks=0,
                        ),
                        self.get_scenario_rename_modal_dialog(scenario_name),
                        dbc.DropdownMenuItem(
                            "Rename",
                            id = {'type':'rename_scenario_mi', 'index': scenario_name},
                            n_clicks=0
                        ),
                        self.get_scenario_duplicate_modal_dialog(scenario_name, scenarios_df),
                        dbc.DropdownMenuItem(
                            "Download",
                            id = {'type':'download_scenario_mi', 'index': scenario_name},
                            n_clicks=0
                        ),
                        dcc.Download(id={'type':'download_scenario_download', 'index': scenario_name}),
                        # html.A(id={'type':'download_scenario_link', 'index': scenario_name},
                        #        children='Download File'
                        #        ),
                        dbc.DropdownMenuItem(divider=True),
                        dbc.DropdownMenuItem(
                            "Delete",
                            id = {'type':'delete_scenario_mi', 'index': scenario_name},
                            n_clicks=0
                        ),
                        self.get_scenario_delete_modal_dialog(scenario_name),
                    ],
                    label="...",
                    size="sm",
                ),
                # html.P(id="item_clicks", className="mt-3"),
            ]
        )
        return dropdown

    def get_scenario_rename_modal_dialog(self, scenario_name: str):
        modal = dbc.Modal(
                    [
                        dbc.ModalHeader("Rename Scenario"),
                        dbc.ModalBody(f"New scenario name for '{scenario_name}'"),
                        dbc.Input(
                            id = {'type':'rename_scenario_modal_input', 'index': scenario_name},
                            value=scenario_name, type="text"),
                        dbc.ModalFooter([
                            dbc.Button("Cancel",
                                       id = {'type':'rename_scenario_modal_cancel', 'index': scenario_name},
                                       className="ml-auto"),
                            dbc.Button("Rename",
                                       id = {'type':'rename_scenario_modal_rename', 'index': scenario_name},
                                       className="ml-auto"),
                        ]),
                    ],
                    id = {'type':'rename_scenario_modal', 'index': scenario_name},
                )
        return modal

    def get_scenario_duplicate_modal_dialog(self, scenario_name: str, scenarios_df=None):
        new_scenario_name = self.dash_app.dbm._find_free_duplicate_scenario_name(scenario_name, scenarios_df)
        modal = dbc.Modal(
            [
                dbc.ModalHeader("Duplicate Scenario"),
                dbc.ModalBody(f"Name for the duplicate of the scenario '{scenario_name}':"),
                dbc.Input(
                    id = {'type':'duplicate_scenario_modal_input', 'index': scenario_name},
                    value=new_scenario_name, type="text"),
                dbc.ModalFooter([
                    dbc.Button("Cancel",
                               id = {'type':'duplicate_scenario_modal_cancel', 'index': scenario_name},
                               className="ml-auto"),
                    dbc.Button("Duplicate",
                               id = {'type':'duplicate_scenario_modal_rename', 'index': scenario_name},
                               className="ml-auto"),
                ]),
            ],
            id = {'type':'duplicate_scenario_modal', 'index': scenario_name},
        )
        return modal

    def get_scenario_delete_modal_dialog(self, scenario_name: str):
        modal = dbc.Modal(
            [
                dbc.ModalHeader("Delete Scenario"),
                dbc.ModalBody(f"Delete the scenario '{scenario_name}':"),
                dbc.ModalFooter([
                    dbc.Button("Cancel",
                               id = {'type':'delete_scenario_modal_cancel', 'index': scenario_name},
                               className="ml-auto"),
                    dbc.Button("Delete",
                               id = {'type':'delete_scenario_modal_rename', 'index': scenario_name},
                               className="ml-auto"),
                ]),
            ],
            id = {'type':'delete_scenario_modal', 'index': scenario_name},
        )
        return modal

    def parse_scenario_upload_contents_callback(self, contents, filename, date):
        """Called for each uploaded scenario"""
        content_type, content_string = contents.split(',')
        print(f"Upload file. filename={filename}, content_type={content_type}")
        decoded = base64.b64decode(content_string)
        # root, file_extension = os.path.splitext(filename)
        file_extension = pathlib.Path(filename).suffix
        scenario_name = pathlib.Path(filename).stem

        print(f"scenario_name = {scenario_name}, extension = {file_extension}")
        try:
            if file_extension == '.xlsx':
                # pass
                # Assume that the user uploaded an excel file
                # df = pd.read_excel(io.BytesIO(decoded))
                xl = pd.ExcelFile(io.BytesIO(decoded))
                # Read data from Excel
                inputs, outputs = ScenarioManager.load_data_from_excel_s(xl)
                # s = f"inputs = {inputs.keys()}, outputs = {outputs.keys()}"
                # print(s)
                print("Input tables: {}".format(", ".join(inputs.keys())))
                print("Output tables: {}".format(", ".join(outputs.keys())))
                self.dash_app.dbm.replace_scenario_in_db(scenario_name=scenario_name, inputs=inputs, outputs=outputs)
                child = html.Div([
                    html.P(f"Uploaded scenario: '{scenario_name}' from '{filename}'"),
                    html.P(f"Input tables: {', '.join(inputs.keys())}"),
                    html.P(f"Output tables: {', '.join(outputs.keys())}"),
                ])
                return child
            elif file_extension == '.zip':
                zip_file = zipfile.ZipFile(io.BytesIO(decoded))
                # unzip_results = [html.P(f"Support for zip archives (of .xslx) is pending: {filename}")]
                unzip_results = []
                for info in zip_file.infolist():
                    scenario_name = pathlib.Path(info.filename).stem
                    file_extension = pathlib.Path(info.filename).suffix
                    if file_extension == '.xlsx':
                        print(f"file in zip : {info.filename}")
                        filecontents = zip_file.read(info)
                        xl = pd.ExcelFile(filecontents)
                        inputs, outputs = ScenarioManager.load_data_from_excel_s(xl)
                        print("Input tables: {}".format(", ".join(inputs.keys())))
                        print("Output tables: {}".format(", ".join(outputs.keys())))
                        self.dash_app.dbm.replace_scenario_in_db(scenario_name=scenario_name, inputs=inputs, outputs=outputs)  #
                        unzip_results.append(html.P(f"Uploaded scenario: '{scenario_name}' from '{info.filename}'"),)
                    else:
                        unzip_results.append(html.P(f"File: '{info.filename}' is not a .xlsx. Skipped."),)
                child = html.Div(unzip_results)
                return child
            else:
                return html.P(f"Unsupported file type: {filename}")
        except Exception as e:
            print(e)
            return html.Div([
                f'There was an error processing this file: {e}'
            ])

        return html.P(f"Uploaded scenario {filename}")

    def set_dash_callbacks(self):
        app = self.dash_app.app

        #############################################################################
        # Scenario operations callbacks
        #############################################################################
        @app.callback(Output('output_data_upload', 'children'),
                      Input('upload_scenario', 'contents'),
                      State('upload_scenario', 'filename'),
                      State('upload_scenario', 'last_modified'))
        def update_output(list_of_contents, list_of_names, list_of_dates):
            """Supports uploading a set of scenarios"""
            if list_of_contents is not None:
                children = [
                    # f"{n}, {d}" for c, n, d in zip(list_of_contents, list_of_names, list_of_dates)
                    self.parse_scenario_upload_contents_callback(c, n, d) for c, n, d in zip(list_of_contents, list_of_names, list_of_dates)
                ]
                return children


        @app.callback([
            Output('download_scenarios_button', 'n_clicks'),
            Output('download_scenarios_download', 'data'),
            ],
            Input('download_scenarios_button', 'n_clicks'),
            prevent_initial_call=True
        )
        def download_scenarios_callback(n_clicks):
            """Download all scenarios in a zip file.
            TODO: download selected set of scenarios
            """
            print("Download all scenarios")
            scenarios_df = self.dash_app.read_scenarios_table_from_db_cached()
            data = None
            with tempfile.TemporaryDirectory() as tmpdir:
                zip_filepath = os.path.join(tmpdir, 'scenarios.zip')
                with zipfile.ZipFile(zip_filepath, 'w') as zipMe:
                    for scenario_name in scenarios_df.index:
                        print(f"Download scenario {scenario_name}")
                        inputs, outputs = self.dash_app.dbm.read_scenario_from_db(scenario_name)
                        filename = f'{scenario_name}.xlsx'
                        filepath = os.path.join(tmpdir, filename)
                        with pd.ExcelWriter(filepath) as writer:
                            ScenarioManager.write_data_to_excel_s(writer, inputs=inputs, outputs=outputs)
                            writer.save()
                            zipMe.write(filepath, arcname=filename, compress_type=zipfile.ZIP_DEFLATED)
                data = dcc.send_file(zip_filepath)

            return 0, data

        @app.callback([
            Output({'type': 'download_scenario_mi', 'index': MATCH}, 'n_clicks'),
            Output({'type': 'download_scenario_download', 'index': MATCH}, 'data'),
            ],
            Input({'type': 'download_scenario_mi', 'index': MATCH}, 'n_clicks'),
            State({'type': 'download_scenario_mi', 'index': MATCH}, 'id'),
            prevent_initial_call=True
            )
        def download_scenario_callback(n_clicks, id):
            """We need the `n_clicks` as input. Only the `id` will not be triggered when a user selects the menu option.
            See https://community.plotly.com/t/excel-writer-to-dcc-download/54132/5 for use of tempfile.TemporaryDirectory()
            """
            scenario_name = id['index']
            print(f"Download scenario {scenario_name}")
            # df = pd.DataFrame({'dropdown_value': [1, 2, 3]})
            # relative_filename = os.path.join(
            #     'downloads',
            #     '{}-download.xlsx'.format(scenario_name)
            # )
            # absolute_filename = os.path.join(os.getcwd(), relative_filename)
            # writer = pd.ExcelWriter(absolute_filename)
            # df.to_excel(writer, 'Sheet1')
            # writer.save()
            # href = './{}'.format(relative_filename)

            multi_threaded = False  # Enabling multi-threading does NOT result in speedup. In fact for small scenarios it is slower!
            inputs, outputs = self.dash_app.dbm.read_scenario_from_db(scenario_name, multi_threaded)
            #TODO: inputs include a scenario table. Remove.

            data = None
            with tempfile.TemporaryDirectory() as tmpdir:
                filename = f'{scenario_name}.xlsx'
                filepath = os.path.join(tmpdir, filename)
                with pd.ExcelWriter(filepath) as writer:
                    ScenarioManager.write_data_to_excel_s(writer, inputs=inputs, outputs=outputs)
                    writer.save()
                    data = dcc.send_file(filepath)

            return 0, data

        # @app.server.route('/downloads/<path:path>')
        # def serve_static(path):
        #     root_dir = os.getcwd()
        #     return flask.send_from_directory(
        #         os.path.join(root_dir, 'downloads'), path
        #     )

        @app.callback(
            Output({'type': 'delete_scenario_modal', 'index': MATCH}, "is_open"),
            [
                Input({'type': 'delete_scenario_mi', 'index': MATCH}, 'n_clicks'),
                Input({'type': 'delete_scenario_modal_cancel', 'index': MATCH}, "n_clicks"),
                Input({'type': 'delete_scenario_modal_rename', 'index': MATCH}, "n_clicks"),
            ],
            [State({'type': 'delete_scenario_modal', 'index': MATCH}, "is_open"),
             # State({'type': 'delete_scenario_modal_input', 'index': MATCH}, "value")
             ],
        )
        def toggle_delete_modal(n1, n2, n3, is_open):
            """

            """
            ctx = dash.callback_context
            if not ctx.triggered:
                raise PreventUpdate

            if n1 or n2 or n3:
                # print(f"Rename modal: {new_scenario_name}")
                # print(f"ctx.triggered[0] = {ctx.triggered[0]}")
                triggered_component_id_str = ctx.triggered[0]['prop_id'].split('.')[0]  # This returns a STRING representation of the pattern-matching id
                # print(f"Rename context id = {triggered_component_id}")
                triggered_component_id_dict = ast.literal_eval(triggered_component_id_str)  # Convert the string to a Dict to get the type.
                ctx_type = triggered_component_id_dict['type']
                current_scenario_name = triggered_component_id_dict['index']
                # print(f"Rename context type = {ctx_type}")

                if ctx_type == 'delete_scenario_modal_rename':
                    print(f"Deleting scenario from {current_scenario_name}")
                    self.dash_app.dbm.delete_scenario_from_db(current_scenario_name)

                return not is_open
            return is_open

        @app.callback(
            Output({'type': 'duplicate_scenario_modal', 'index': MATCH}, "is_open"),
            [
                Input({'type': 'duplicate_scenario_mi', 'index': MATCH}, 'n_clicks'),
                Input({'type': 'duplicate_scenario_modal_cancel', 'index': MATCH}, "n_clicks"),
                Input({'type': 'duplicate_scenario_modal_rename', 'index': MATCH}, "n_clicks"),
            ],
            [State({'type': 'duplicate_scenario_modal', 'index': MATCH}, "is_open"),
             State({'type': 'duplicate_scenario_modal_input', 'index': MATCH}, "value")
             ],
        )
        def toggle_duplicate_modal(n1, n2, n3, is_open, new_scenario_name):
            """
            TODO: replace by a duplicate + delete
            """
            ctx = dash.callback_context
            if not ctx.triggered:
                raise PreventUpdate

            if n1 or n2 or n3:
                # print(f"Rename modal: {new_scenario_name}")
                # print(f"ctx.triggered[0] = {ctx.triggered[0]}")
                triggered_component_id_str = ctx.triggered[0]['prop_id'].split('.')[0]  # This returns a STRING representation of the pattern-matching id
                # print(f"Rename context id = {triggered_component_id}")
                triggered_component_id_dict = ast.literal_eval(triggered_component_id_str)  # Convert the string to a Dict to get the type.
                ctx_type = triggered_component_id_dict['type']
                current_scenario_name = triggered_component_id_dict['index']
                # print(f"Rename context type = {ctx_type}")

                if ctx_type == 'duplicate_scenario_modal_rename':
                    if new_scenario_name != current_scenario_name:
                        print(f"Duplicating scenario from {current_scenario_name} to {new_scenario_name}")
                        self.dash_app.dbm.duplicate_scenario_in_db(current_scenario_name, new_scenario_name)

                return not is_open
            return is_open

        @app.callback(
            Output({'type': 'rename_scenario_modal', 'index': MATCH}, "is_open"),
            [
             Input({'type': 'rename_scenario_mi', 'index': MATCH}, 'n_clicks'),
             Input({'type': 'rename_scenario_modal_cancel', 'index': MATCH}, "n_clicks"),
             Input({'type': 'rename_scenario_modal_rename', 'index': MATCH}, "n_clicks"),
             ],
            [State({'type': 'rename_scenario_modal', 'index': MATCH}, "is_open"),
             State({'type': 'rename_scenario_modal_input', 'index': MATCH}, "value")
             ],
                )
        def toggle_rename_modal(n1, n2, n3, is_open, new_scenario_name):
            """
            TODO: trigger reload of scenario table and update of UI
            If `rename_scenario_modal_rename` then do the rename
            """
            ctx = dash.callback_context
            if not ctx.triggered:
                raise PreventUpdate

            if n1 or n2 or n3:
                # print(f"Rename modal: {new_scenario_name}")
                # print(f"ctx.triggered[0] = {ctx.triggered[0]}")
                triggered_component_id_str = ctx.triggered[0]['prop_id'].split('.')[0]  # This returns a STRING representation of the pattern-matching id
                # print(f"Rename context id = {triggered_component_id}")
                triggered_component_id_dict = ast.literal_eval(triggered_component_id_str)  # Convert the string to a Dict to get the type.
                ctx_type = triggered_component_id_dict['type']
                current_scenario_name = triggered_component_id_dict['index']
                # print(f"Rename context type = {ctx_type}")

                if ctx_type == 'rename_scenario_modal_rename':
                    if new_scenario_name != current_scenario_name:
                        print(f"Renaming scenario from {current_scenario_name} to {new_scenario_name}")
                        self.dash_app.dbm.rename_scenario_in_db(current_scenario_name, new_scenario_name)

                return not is_open
            return is_open

        # @app.callback(
        #     Output({'type': 'duplicate_scenario_mi', 'index': MATCH}, 'n_clicks'),
        #     Input({'type': 'duplicate_scenario_mi', 'index': MATCH}, 'n_clicks'),
        #     State({'type': 'duplicate_scenario_mi', 'index': MATCH}, 'id'),
        #     prevent_initial_call=True
        # )
        # def duplicate_scenario_callback(n_clicks, id):
        #     """We need the `n_clicks` as input. Only the `id` will not be triggered when a user selects the menu option."""
        #     scenario_name = id['index']
        #     print(f"Duplicate scenario {scenario_name}")
        #     raise PreventUpdate
        #     return 0
        #
        # @app.callback(
        #     Output({'type': 'rename_scenario_mi', 'index': MATCH}, 'n_clicks'),
        #     Input({'type': 'rename_scenario_mi', 'index': MATCH}, 'n_clicks'),
        #     State({'type': 'rename_scenario_mi', 'index': MATCH}, 'id'),
        #     prevent_initial_call=True
        # )
        # def rename_scenario_callback(n_clicks, id):
        #     """We need the `n_clicks` as input. Only the `id` will not be triggered when a user selects the menu option."""
        #     scenario_name = id['index']
        #     print(f"Rename scenario {scenario_name}")
        #     raise PreventUpdate
        #     return 0
        #
        # @app.callback(
        #     Output({'type': 'delete_scenario_mi', 'index': MATCH}, 'n_clicks'),
        #     Input({'type': 'delete_scenario_mi', 'index': MATCH}, 'n_clicks'),
        #     State({'type': 'delete_scenario_mi', 'index': MATCH}, 'id'),
        #     prevent_initial_call=True
        # )
        # def delete_scenario_callback(n_clicks, id):
        #     """We need the `n_clicks` as input. Only the `id` will not be triggered when a user selects the menu option."""
        #     scenario_name = id['index']
        #     print(f"Delete scenario {scenario_name}")
        #     raise PreventUpdate
        #     return 0

        #############################################################################
        @app.callback(
            Output("collapse_stop_server_button_state", "is_open"),
            [Input("collapse_stop_server_button", "n_clicks")],
            [State("collapse_stop_server_button_state", "is_open")],
        )
        def toggle_collapse(n, is_open):
            if n:
                return not is_open
            return is_open

        @app.callback(
            Output('stop_server_button_output', 'children'),
            Input('stop_server_button', 'n_clicks'),
            prevent_initial_call=True
        )
        def update_output(n_clicks):
            self.dash_app.shutdown()
            return 'The stop button has been clicked {} times.'.format(n_clicks)

        ##############################################################################
        #  Select Reference Scenarios
        ##############################################################################
        @app.callback(
            Output('reference_scenario_name_store', 'data'),
            Input('reference_scenario_drpdwn', 'value'),
        )
        def store_reference_scenario(reference_scenario_name):
            # print(f"Selected ref scenario = {reference_scenario_name}")
            return reference_scenario_name

        @app.callback(
            Output('multi_scenario_names_store', 'data'),
            Input('reference_scenarios_checklist', 'value'),
        )
        def store_multi_scenario_names(multi_scenario_names):
            # print(f"Multi scenario names = {multi_scenario_names}")
            return multi_scenario_names